<?php

namespace June\Components\Api\Client;

use GuzzleHttp\Exception\GuzzleException;
use GuzzleHttp\Exception\TransferException;
use GuzzleHttp\Exception\RequestException;
use GuzzleHttp\Exception\BadResponseException;
use GuzzleHttp\Exception\ClientException;   // 400-
use GuzzleHttp\Exception\ServerException;   // 500-
use June\Components\Logger\Helper\LoggerHelper;
use Shopware\Core\System\SystemConfig\SystemConfigService;

class JuneApiClient {

    const JUNE_TRANSFER_TYPE = "June.config.transfertype";
    const DEFAULT_TIMEOUT = 4;

    /**
     * @var SystemConfigService
     */
    private SystemConfigService $configService;

    private LoggerHelper $loggerHelper;

    public function __construct
    (
        SystemConfigService $configService,
        LoggerHelper $loggerHelper
    )
    {
        $this->configService = $configService;
        $this->loggerHelper = $loggerHelper;
    }

    public function sendData($data, $url, $auth=null, $timeout=self::DEFAULT_TIMEOUT) {
        if (!$timeout || $timeout<0) {
            $timeout = self::DEFAULT_TIMEOUT;
        }

        if (!empty($data)) {
            if ($this->configService->get(self::JUNE_TRANSFER_TYPE)) {
                $this->socketRequest($data, $url);
            } else {
                return $this->httpRequest($data, $url, $auth, $timeout);
            }
        }

        return null;
    }

    private function socketRequest($data, $url) {
        $endpoint = $url;
        $postData = json_encode($data);
        $endpointParts = parse_url($endpoint);
        $endpointParts['path'] = $endpointParts['path'] ?? '/';
        $endpointParts['port'] = $endpointParts['port'] ?? $endpointParts['scheme'] === 'https' ? 443 : 80;

        $contentLength = strlen($postData);

        $request = "POST {$endpointParts['path']} HTTP/1.1\r\n";
        $request .= "Host: {$endpointParts['host']}\r\n";
        $request .= "Content-Type: application/json\r\n";
        $request .= "Content-Length: {$contentLength}\r\n";
        $request .= "Connection: Close\r\n\r\n";
        $request .= $postData;

        $prefix = substr($endpoint, 0, 8) === 'https://' ? 'tls://' : ''; // maybe use ssl?


        $this->loggerHelper->logSocket("--- Step 1: Connection Info ---");
        $this->loggerHelper->logSocket("Endpoint: ".$prefix.$endpointParts['host']);
        $this->loggerHelper->logSocket("Port: ".$endpointParts['port']);

        try {
            $this->loggerHelper->logSocket("--- Step 2: Start Connection ---");
            $socket = fsockopen($prefix.$endpointParts['host'], $endpointParts['port'], $errNo, $errStr, 1);

            if(!$socket || $errNo || $errStr) {
                $this->loggerHelper->logSocket("Connection Failed {$errNo}: {$errStr}");
            } else {
                $this->loggerHelper->logSocket("Connection is established");

                fwrite($socket, $request);

                if(fflush($socket)) {
                    $this->loggerHelper->logSocket("POST Data Success");
                } else {
                    $this->loggerHelper->logSocket("POST Data Failed");
                }
            }
            fclose($socket);
            $this->loggerHelper->logSocket("--- Step 3: Close Connection ---");

        } catch(\Exception $ex) {
            $this->loggerHelper->logSocket("ERROR: ".$ex->getMessage());
        }
    }

    private function httpRequest($data, $url, $auth, $timeout, $method='POST') {
        $header = ['Accept' => 'application/json'];
        if ($auth != null) {
            $header['Authorization'] = 'Bearer ' . $auth;
        }
        $options["headers"] = $header;
        $options['json'] = $data;
        $client = new \GuzzleHttp\Client(["timeout" => $timeout]);

        try {
            switch (strtolower($method)) {
                case 'post':
                    return $client->post($url, $options);

                case 'put';
                    return $client->put($url, $options);

                case 'get':
                    return $client->get($url, $options);

            }
        } catch (ClientException $e) {
            // return response so httpStatus can be checked to decide on further actions
            return $e->getResponse();
        } catch (TransferException $e) {
            $errorMessage = $e->getMessage();
            $timeoutException = false;
            if (str_starts_with($errorMessage, "cURL error 28: Operation timed out")) {
                $timeoutException = true;
            }
            if (!$timeoutException || $timeout!==self::DEFAULT_TIMEOUT) {
                $this->loggerHelper->logError($errorMessage, 'httpRequest');
            }
        }

        return null;
    }


    public function upsertCustomer($customer, $collectId, $collectToken, $collectAuth, $timeout=null) {
        if (!$timeout || $timeout<0) {
            $timeout = self::DEFAULT_TIMEOUT;
        }

        if ($collectId) {
            $updateResponse = $this->updateCollect($collectId, $customer, $collectToken, $collectAuth, $timeout);
            if ($updateResponse && $updateResponse->getStatusCode()===200) {
                return json_decode($updateResponse->getBody()->getContents(), true);
            } else {
                //@ignore: Guzzle error will already be shown
                // $this->loggerHelper->logError("Could not update customer to June: ".var_export($customer, 1), 'httpRequest');
            }
        } else {
            $createResponse = $this->httpRequest($customer, "https://engagement.juneapp.com/v2/collection/$collectToken/collects", $collectAuth, $timeout, 'POST');

            if ($createResponse) {
                switch ($createResponse->getStatusCode()) {
                    // created
                    case 200:
                        return json_decode($createResponse->getBody()->getContents(), true);

                    // already exists
                    case 406:
                        $getResponse = $this->getCollectByEmail($customer['email'], $collectToken, $collectAuth);
                        if ($getResponse && $getResponse->getStatusCode() === 200) {
                            $getBody = json_decode($getResponse->getBody()->getContents(), true);

                            $updateResponse = $this->updateCollect($getBody['id'], $customer, $collectToken, $collectAuth, $timeout);
                            if ($updateResponse && $updateResponse->getStatusCode() === 200) {
                                return json_decode($updateResponse->getBody()->getContents(), true);
                            } else {
                                //@ignore: Guzzle error will already be shown
                                // $this->loggerHelper->logError("Could not update customer to June: " . var_export($customer, 1), 'httpRequest');
                            }
                        }
                        break;

                    default:
                        //@ignore: Guzzle error will already be shown
                        // $this->loggerHelper->logError("Could not upsert customer to June: " . var_export($customer, 1), 'httpRequest');
                }
            } else {
                //@ignore: Guzzle error will already be shown
                // $this->loggerHelper->logError("Could not upsert customer to June: " . var_export($customer, 1), 'httpRequest');
            }
        }

        return null;
    }

    private function getCollectByEmail($email, $collectToken, $collectAuth, $timeout=null) {
        $email = urlencode($email);
        return $this->httpRequest(null, "https://engagement.juneapp.com/v2/collection/$collectToken/collects/search-by/email/$email", $collectAuth, $timeout, 'GET');
    }

    private function updateCollect($collectId, $customer, $collectToken, $collectAuth, $timeout=null) {
        return $this->httpRequest($customer, "https://engagement.juneapp.com/v2/collection/$collectToken/collects/$collectId", $collectAuth, $timeout, 'PUT');
    }
}
