<?php declare(strict_types=1);

namespace June\Core\Content\Flow\Dispatching\Action;

use Shopware\Core\Content\Flow\Dispatching\Action\FlowAction;
use Shopware\Core\Content\Flow\Dispatching\Action\FlowMailVariables;
use Shopware\Core\Content\Flow\Dispatching\StorableFlow;
use Shopware\Core\Framework\DataAbstractionLayer\EntityRepository;
use June\Components\Api\Client\JuneApiClient;
use June\Components\Logger\Helper\LoggerHelper;
use June\Core\Framework\Event\JuneMailAware;
use Shopware\Core\Framework\Event\MailAware;

use Shopware\Core\Content\MailTemplate\Exception\MailEventConfigurationException;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Criteria;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Filter\EqualsFilter;
use Shopware\Core\System\SystemConfig\SystemConfigService;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;


class JuneMailAction extends FlowAction implements EventSubscriberInterface
{
    const JUNE_API_URL = 'June.config.juneapiurl';
    const JUNE_API_AUTH_TOKEN = 'June.config.juneauthtoken';
    const JUNE_TRANSFER_GUEST_DATA = 'June.config.transferFlowGuestData';

    /**
     * @var SystemConfigService
     */
    private SystemConfigService $configService;

    private EntityRepository $juneFlowActionRepository;

    private LoggerHelper $loggerHelper;

    private JuneApiClient $juneApiClient;

    public function __construct
    (
        SystemConfigService $configService,
        EntityRepository $juneFlowActionRepository,
        LoggerHelper $loggerHelper,
        JuneApiClient $juneApiClient
    )
    {
        $this->configService = $configService;
        $this->juneFlowActionRepository = $juneFlowActionRepository;
        $this->loggerHelper = $loggerHelper;
        $this->juneApiClient = $juneApiClient;
    }

    public static function getName(): string
    {
        // your own action name
        return 'action.june.mail';
    }

    public static function getSubscribedEvents(): array
    {
        return [
            self::getName() => 'handleFlow',
        ];
    }

    public function requirements(): array
    {
        return [MailAware::class];
    }

    public function handleFlow(StorableFlow $flow): void
    {
        $config = $flow->getConfig();
        $eventName = $flow->getName();

        if (!$flow->hasData(MailAware::MAIL_STRUCT)) {
            $this->loggerHelper->logError("Can't access mailStruct information for MailFlow.: {$eventName}", 'juneFlowActionConfig');
            return;
        }

        $mailStruct = $flow->getData(MailAware::MAIL_STRUCT);
        $recipients = $mailStruct->getRecipients();

        if (empty($recipients)) {
            $this->loggerHelper->logError("No recipients found for MailFlow.: {$eventName}", 'juneFlowActionConfig');
            return;
        }

        $contentData = $flow->data();

        try {
            $customerGuest = (\array_key_exists('order', $contentData) && $contentData['order']->getOrderCustomer() ? $contentData['order']->getOrderCustomer()->getCustomer()->getGuest() : null);
        } catch (\Exception $e) {
            $customerGuest = null;
            $this->loggerHelper->logError($e->getMessage(), 'juneFlowActionConfig');
        }

        if(
            $this->configService->get(self::JUNE_TRANSFER_GUEST_DATA)
            || !$customerGuest
        ) {
            try {
                // define base information for JUNE
                $returnData = [
                    "action_type"   => $eventName,
                    "store_data"    => [
                        "sales_channel_id"  => $flow->getData('salesChannelId'),
                        "shop_name"         => $flow->getData(FlowMailVariables::SHOP_NAME),
                    ],
                    "user_data"     => [],
                    "content_data"  => []
                ];

                foreach($recipients as $email => $name) {
                    $returnData["user_data"] = [
                        'email' => $email,
                        'name' => $name
                    ];
                }
                $returnData["email"] = $returnData["recipient_email"] = $returnData["user_data"]["email"];

                foreach($contentData as $keyData => $data) {
                    $returnData["content_data"][$keyData] = $data;
                }

                if(str_contains($eventName, "newsletter")) {
                    $returnData["user_data"] = [
                        "firstname" => $contentData['newsletterRecipient']->getFirstName(),
                        "lastname"  => $contentData['newsletterRecipient']->getLastName()
                    ];

                    $returnData["content_data"] = [
                        "intent" => ($eventName === "newsletter.unsubscribe" ? "unsubscribe" : "subscribe")
                    ];
                } else if($eventName == "checkout.order.placed") {
                    foreach($contentData as $keyData => $data) {
                        $returnData["content_data"][$keyData] = $data;

                        if ($data->addresses && $data->billingAddressId) {
                            $returnData["content_data"][$keyData]["billing"] = [];
                            foreach ($data->addresses as $address) {
                                if($address->id == $data->billingAddressId) {
                                    $returnData["content_data"][$keyData]["billing"] = [
                                        'firstname' => $address->firstName,
                                        'lastname' => $address->lastName,
                                        'street' => $address->street,
                                        'zipcode' => $address->zipcode,
                                        'city' => $address->city,
                                        "country" =>$address->country->name
                                    ];
                                }
                            }
                        }

                        if ($data->deliveries) {
                            $returnData["content_data"][$keyData]["shipping"] = [];
                            foreach ($data->deliveries as $shippingAddress) {
                                if ($shippingAddress->shippingOrderAddress) {
                                    $returnData["content_data"][$keyData]["shipping"][] = [
                                        "firstname" => $shippingAddress->shippingOrderAddress->firstName,
                                        "lastname" => $shippingAddress->shippingOrderAddress->lastName,
                                        "street" => $shippingAddress->shippingOrderAddress->street,
                                        "zipcode" => $shippingAddress->shippingOrderAddress->zipcode,
                                        "city" => $shippingAddress->shippingOrderAddress->city,
                                        "country" => $shippingAddress->shippingOrderAddress->country->name
                                    ];
                                }
                            }
                        }
                    }
                }

                $url = $this->configService->get(self::JUNE_API_URL);
                $authToken = $this->configService->get(self::JUNE_API_AUTH_TOKEN);

                if($config['juneFlowAction'] != NULL) {
                    $context = $flow->getContext();
                    $criteria = new Criteria();

                    $criteria->addFilter(
                        new EqualsFilter('id', $config['juneFlowAction'] )
                    );
                    $result = $this->juneFlowActionRepository->search($criteria, $context)->first();

                    $collectToken = $result->getCollectToken();
                } else {
                    $this->loggerHelper->logError("There is no Collection configured in the FlowAction: {$eventName}", 'juneFlowActionConfig');
                    return;
                }

                $this->juneApiClient->sendData($returnData, $url.$collectToken, $authToken);

            } catch (\Exception $e) {
                $this->loggerHelper->logError($e->getMessage(), 'juneFlowAction');
            }
        }
    }

    /*private function buildAttachments(JuneMailAware $mailEvent, MailTemplateEntity $mailTemplate, MailSendSubscriberConfig $extensions, array $eventConfig): array
    {
        $attachments = [];

        if ($mailTemplate->getMedia() !== null) {
            foreach ($mailTemplate->getMedia() as $mailTemplateMedia) {
                if ($mailTemplateMedia->getMedia() === null) {
                    continue;
                }
                if ($mailTemplateMedia->getLanguageId() !== null && $mailTemplateMedia->getLanguageId() !== $mailEvent->getContext()->getLanguageId()) {
                    continue;
                }

                $attachments[] = $this->mediaService->getAttachment(
                    $mailTemplateMedia->getMedia(),
                    $mailEvent->getContext()
                );
            }
        }

        if (!empty($extensions->getMediaIds())) {
            $criteria = new Criteria($extensions->getMediaIds());
            $criteria->setTitle('send-mail::load-media');

            $entities = $this->mediaRepository->search($criteria, $mailEvent->getContext());

            foreach ($entities as $media) {
                $attachments[] = $this->mediaService->getAttachment($media, $mailEvent->getContext());
            }
        }

        if (!empty($extensions->getDocumentIds())) {
            $attachments = $this->buildOrderAttachments($extensions->getDocumentIds(), $attachments, $mailEvent->getContext());
        }

        if (empty($eventConfig['documentTypeIds']) || !\is_array($eventConfig['documentTypeIds']) || !$mailEvent instanceof OrderAware) {
            return $attachments;
        }

        return $this->buildFlowSettingAttachments($mailEvent->getOrderId(), $eventConfig['documentTypeIds'], $attachments, $mailEvent->getContext());
    }*/

    /*private function injectTranslator(JuneMailAware $event): bool
    {
        if ($event->getSalesChannelId() === null) {
            return false;
        }

        if ($this->translator->getSnippetSetId() !== null) {
            return false;
        }

        $this->translator->injectSettings(
            $event->getSalesChannelId(),
            $event->getContext()->getLanguageId(),
            $this->languageLocaleProvider->getLocaleForLanguageId($event->getContext()->getLanguageId()),
            $event->getContext()
        );

        return true;
    }*/

    /*private function buildOrderAttachments(array $documentIds, array $attachments, Context $context): array
    {
        $criteria = new Criteria($documentIds);
        $criteria->setTitle('send-mail::load-attachments');
        $criteria->addAssociation('documentMediaFile');
        $criteria->addAssociation('documentType');

        $entities = $this->documentRepository->search($criteria, $context);

        return $this->mappingAttachmentsInfo($entities, $attachments, $context);
    }*/

    /*private function buildFlowSettingAttachments(string $orderId, array $documentTypeIds, array $attachments, Context $context): array
    {
        $documents = $this->connection->fetchAllAssociative(
            'SELECT
                LOWER(hex(`document`.`document_type_id`)) as doc_type,
                LOWER(hex(`document`.`id`)) as doc_id,
                `document`.`created_at` as newest_date
            FROM
                `document`
            WHERE
                HEX(`document`.`order_id`) = :orderId
                AND HEX(`document`.`document_type_id`) IN (:documentTypeIds)
            ORDER BY `document`.`created_at` DESC',
            [
                'orderId' => $orderId,
                'documentTypeIds' => $documentTypeIds,
            ],
            [
                'documentTypeIds' => Connection::PARAM_STR_ARRAY,
            ]
        );

        $documentsGroupByType = FetchModeHelper::group($documents);

        foreach ($documentsGroupByType as $document) {
            $documentIds[] = array_shift($document)['doc_id'];
        }

        if (empty($documentIds)) {
            return $attachments;
        }

        $criteria = new Criteria($documentIds);
        $criteria->setTitle('send-mail::load-flow-documents');
        $criteria->addAssociations(['documentMediaFile', 'documentType']);

        $entities = $this->documentRepository->search($criteria, $context);

        return $this->mappingAttachmentsInfo($entities, $attachments, $context);
    }*/

    /*private function mappingAttachmentsInfo(EntityCollection $entities, array $attachments, Context $context): array
    {
        foreach ($entities as $document) {
            $documentId = $document->getId();
            $document = $this->documentService->getDocument($document, $context);

            $attachments[] = [
                'id' => $documentId,
                'content' => $document->getFileBlob(),
                'fileName' => $document->getFilename(),
                'mimeType' => $document->getContentType(),
            ];
        }

        return $attachments;
    }*/
}
