<?php declare(strict_types=1);

namespace June\Service\ScheduledTask;

use June\Components\Api\Client\JuneApiClient;
use June\Components\Logger\Helper\LoggerHelper;
use Shopware\Core\Checkout\Customer\CustomerEntity;
use Shopware\Core\Framework\Context;
use Shopware\Core\Framework\DataAbstractionLayer\EntityRepository;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Criteria;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Filter\EqualsFilter;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Filter\RangeFilter;
use Shopware\Core\Framework\MessageQueue\ScheduledTask\ScheduledTask;
use Shopware\Core\Framework\MessageQueue\ScheduledTask\ScheduledTaskDefinition;
use Shopware\Core\Framework\MessageQueue\ScheduledTask\ScheduledTaskEntity;
use Shopware\Core\Framework\MessageQueue\ScheduledTask\ScheduledTaskHandler;
use Shopware\Core\System\SystemConfig\SystemConfigService;
use Shopware\Core\Content\Newsletter\Aggregate\NewsletterRecipient\NewsletterRecipientEntity;
use June\Subscriber\JuneCommerceSubscriber;

use Shopware\Core\System\Locale\LocaleEntity;
use Shopware\Core\System\Country\CountryEntity;
use Shopware\Core\System\Tag\TagEntity;

#[AsMessageHandler(handles: ExportNewsletterSubscriberTask::class)]
class ExportNewsletterSubscriberTaskHandler extends ScheduledTaskHandler {

    const SUBSCRIBER_LIMIT_PER_REQUEST = 100;

    private EntityRepository $pluginRepository;

    private EntityRepository $languageRepository;

    private EntityRepository $localeRepository;

    private EntityRepository $countryRepository;

    private EntityRepository $newsletterSubscriberRepository;

    private EntityRepository $customerRepository;

    private EntityRepository $customerTagsRepository;

    private EntityRepository $salesChannelRepository;

    private SystemConfigService $configService;

    private LoggerHelper $loggerHelper;

    private JuneApiClient $juneApiClient;

    private array $loadedSubscriberTags = [];

    public function __construct
    (
        EntityRepository $scheduledTaskRepository,
        EntityRepository $pluginRepository,
        EntityRepository $languageRepository,
        EntityRepository $localeRepository,
        EntityRepository $countryRepository,
        EntityRepository $newsletterSubscriberRepository,
        EntityRepository $customerRepository,
        EntityRepository $customerTagsRepository,
        EntityRepository $salesChannelRepository,
        SystemConfigService $configService,
        LoggerHelper $loggerHelper,
        JuneApiClient $juneApiClient
    )
    {
        parent::__construct($scheduledTaskRepository);
        $this->pluginRepository = $pluginRepository;
        $this->languageRepository = $languageRepository;
        $this->localeRepository = $localeRepository;
        $this->countryRepository = $countryRepository;
        $this->newsletterSubscriberRepository = $newsletterSubscriberRepository;
        $this->customerRepository = $customerRepository;
        $this->customerTagsRepository = $customerTagsRepository;
        $this->salesChannelRepository = $salesChannelRepository;
        $this->configService = $configService;
        $this->loggerHelper = $loggerHelper;
        $this->juneApiClient = $juneApiClient;
    }

    /**
     * @param ScheduledTask $task
     * @return void
     */
    public function __invoke(ScheduledTask $task): void
    {
        $taskId = $task->getTaskId();
        if ($taskId === null) {
            // run task independent of the schedule
            $this->run();

            return;
        }

        /** @var ScheduledTaskEntity|null $taskEntity */
        $taskEntity = $this->scheduledTaskRepository
            ->search(new Criteria([$taskId]), Context::createDefaultContext())
            ->get($taskId);

        if ($taskEntity === null || !$taskEntity->isExecutionAllowed()) {
            return;
        }

        $this->markTaskRunning($task);

        try {
            $this->run();
        } catch (\Throwable $e) {
            $this->markTaskFailed($task);

            $this->loggerHelper->logError($e->getMessage(), "exportSubscribers");
        }

        if($this->configService->get('June.config.startSubscriberExport') === true) {
            $subscribers = $this->getNewsletterSubscribers();
            $customers = $this->getCustomers();
            if (count($subscribers) > 0 || count($customers)>0) {
                $this->rescheduleTask($task, $taskEntity);
            } else {
                $this->scheduledTaskRepository->update([
                    [
                        'id' => $task->getTaskId(),
                        'status' => ScheduledTaskDefinition::STATUS_INACTIVE,
                    ],
                ], Context::createDefaultContext());

                $this->loggerHelper->logInfo("Subscriber export finished for a total of ".$this->getTotalNewsletterSubscribersCount()." newsletter subscribers and ".$this->getTotalCustomersCount()." customers.", "exportSubscribers");
            }
        } else {
            $this->loggerHelper->logInfo("Subscriber export inactive.", "exportSubscribers");

            $this->scheduledTaskRepository->update([
                [
                    'id' => $task->getTaskId(),
                    'status' => ScheduledTaskDefinition::STATUS_INACTIVE,
                ],
            ], Context::createDefaultContext());
        }
    }

    /**
     * @return iterable
     */
    public static function getHandledMessages(): iterable
    {
        return [ExportNewsletterSubscriberTask::class];
    }

    /**
     * @return void
     */
    public function run(): void
    {
        $commerceCollectToken = $this->configService->get(JuneCommerceSubscriber::JUNE_COMMERCE_COLLECT_TOKEN);
        $commerceCollectAuth = $this->configService->get(JuneCommerceSubscriber::JUNE_COMMERCE_COLLECT_AUTH);

        if($this->configService->get('June.config.startSubscriberExport') === true && !empty($commerceCollectToken) && !empty($commerceCollectAuth)) {
            try {
                $customerData = [];
                $customerUpdateFields = [
                    "email",
                    "firstname",
                    "lastname",
                    "salutation",
                    "salutation_readable",
                    "locale",
                    "updated_at",
                    "guest"
                ];

                $users = $this->getCustomers();
                $isNewsletterEntities = false;
                if (empty($users)) {
                    $users = $this->getNewsletterSubscribers();
                    $isNewsletterEntities = true;
                }

                if ($users) {
                    foreach ($users as $user) {
                        if (!$user || !$user->getId() || !$user->getEmail() || isset($customerData[$user->getEmail()])) {
                            continue;
                        }

                        if ($isNewsletterEntities) {
                            $subscriber = $user;
                            $customer = $this->getCustomer($subscriber);
                        } else {
                            $customer = $user;
                            $subscriber = $this->getNewsletterSubscriber($customer);
                        }

                        $locale = $this->getLocale($subscriber ?? $customer);
                        $country = $this->getCountry($subscriber ?? $customer);
                        $email = $user->getEmail();

                        $enrichedCustomer = [
                            'customer_id'           => $customer?->getCustomerNumber() ?? $subscriber?->getId(),
                            'email'                 => $email,
                            'firstname'             => $customer?->getFirstname() ?? $subscriber?->getFirstname(),
                            'lastname'              => $customer?->getLastname() ?? $subscriber?->getLastname(),
                            'salutation'            => $customer?->getSalutation()?->getSalutationKey() ?? $subscriber?->getSalutation()?->getSalutationKey(),
                            'salutation_readable'   => $customer?->getSalutation()?->getLetterName() ?? $subscriber?->getSalutation()?->getLetterName(),
                            'store_id'              => $subscriber?->getSalesChannelId() ?? $customer?->getSalesChannelId(),
                            'locale'                => $locale?->getCode(),
                            'geo_location'          => [
                                'country'               => $country?->getName(),
                            ],
                            'guest'                 => !$customer || $customer->getGuest(),
                            'created_at'            => $user->getCreatedAt()->format('c'),
                            'updated_at'            => $user->getUpdatedAt()?->format('c'),
                        ];

                        if ($subscriber?->getConfirmedAt()) {
                            $enrichedCustomer['subscription_confirmed_at'] = $subscriber->getConfirmedAt()->format('c');
                            $customerUpdateFields[] = 'subscription_confirmed_at';
                        }

                        // add more information to customer if available
                        if ($customer?->getCustomFields()) {
                            foreach ($customer->getCustomFields() as $fieldName => $fieldValue) {
                                if (!isset($enrichedCustomer[$fieldName])) {
                                    $transformedFieldName = ($enrichedCustomer['guest'] ? 'guest_' : '') . $fieldName;
                                    $enrichedCustomer[$transformedFieldName] = $fieldValue;

                                    if (!in_array($transformedFieldName, $customerUpdateFields)) {
                                        $customerUpdateFields[] = $transformedFieldName;
                                    }
                                }
                            }
                        }
                        if ($customer?->getBirthday()) {
                            $enrichedCustomer['birthday'] = $customer->getBirthday()->format('c');
                            if (!in_array('birthday', $customerUpdateFields)) {
                                $customerUpdateFields[] = 'birthday';
                            }
                        }
                        if ($customer?->getLastLogin()) {
                            $enrichedCustomer['last_login'] = $customer->getLastLogin()->format('c');
                            if (!in_array('last_login', $customerUpdateFields)) {
                                $customerUpdateFields[] = 'last_login';
                            }
                        }
                        if ($customer?->getLastOrderDate()) {
                            $enrichedCustomer['last_order_date'] = $customer->getLastOrderDate()->format('c');
                            if (!in_array('last_order_date', $customerUpdateFields)) {
                                $customerUpdateFields[] = 'last_order_date';
                            }
                        }
                        if ($customer?->getOrderCount()) {
                            $enrichedCustomer['order_count'] = $customer->getOrderCount();
                            if (!in_array('order_count', $customerUpdateFields)) {
                                $customerUpdateFields[] = 'order_count';
                            }
                        }
                        if ($customer?->getOrderTotalAmount()) {
                            $enrichedCustomer['order_total_amount'] = $customer->getOrderTotalAmount();
                            if (!in_array('order_total_amount', $customerUpdateFields)) {
                                $customerUpdateFields[] = 'order_total_amount';
                            }
                        }

                        if ($customer?->getCampaignCode()) {
                            $enrichedCustomer['campaign_code'] = $customer->getCampaignCode();
                            if (!in_array('campaign_code', $customerUpdateFields)) {
                                $customerUpdateFields[] = 'campaign_code';
                            }
                        }

                        if ($customer?->getAffiliateCode()) {
                            $enrichedCustomer['affiliate_code'] = $customer->getAffiliateCode();
                            if (!in_array('affiliate_code', $customerUpdateFields)) {
                                $customerUpdateFields[] = 'affiliate_code';
                            }
                        }

                        $enrichedCustomer['_unsubscribed'] = !($subscriber?->getStatus() === "direct" || $subscriber?->getStatus() === "optIn");
                        if (!in_array('_unsubscribed', $customerUpdateFields)) {
                            $customerUpdateFields[] = '_unsubscribed';
                        }

                        $enrichedCustomer['intent'] = $subscriber?->getStatus();
                        if (!in_array('intent', $customerUpdateFields)) {
                            $customerUpdateFields[] = 'intent';
                        }

                        if ($subscriber?->getCity()) {
                            $enrichedCustomer['city'] = $subscriber->getCity();
                            $enrichedCustomer['geo_location']['city'] = $subscriber->getCity();
                            if (!in_array('city', $customerUpdateFields)) {
                                $customerUpdateFields[] = 'city';
                            }
                        }
                        if ($subscriber?->getZipCode()) {
                            $enrichedCustomer['zip_code'] = $subscriber->getZipCode();
                            if (!in_array('zip_code', $customerUpdateFields)) {
                                $customerUpdateFields[] = 'zip_code';
                            }
                        }


                        if ($customer?->getTagIds()) {
                            $enrichedCustomer['_tags'] = [];
                            foreach ($customer->getTags() as $tag) {
                                if ($tag?->getName() && !in_array($tag->getName(), $enrichedCustomer['_tags'])) {
                                    $enrichedCustomer['_tags'][] = $tag->getName();
                                }
                            }
                        }
                        if ($subscriber?->getTags()) {
                            if (!isset($enrichedCustomer['_tags'])) {
                                $enrichedCustomer['_tags'] = [];
                            }
                            foreach ($subscriber->getTags() as $tag) {
                                if ($tag?->getName() && !in_array($tag->getName(), $enrichedCustomer['_tags'])) {
                                    $enrichedCustomer['_tags'][] = $tag->getName();
                                }
                            }
                        }

                        $customerData[$email] = $enrichedCustomer;
                    }

                    // send collected data to June
                    if (!empty($customerData)) {
                        $requestData = [
                            "json" => array_values($customerData),
                            "activate_trigger" => true,
                            "default_subscribe" => false,
                            "force_subscribe" => false,
                            "update_if_exists" => true,
                            "update_fields" => $customerUpdateFields,
                        ];

                        $this->juneApiClient->sendData($requestData, "https://engagement.juneapp.com/collect/$commerceCollectToken/import/json", $commerceCollectAuth, 6);
                    }

                    if ($isNewsletterEntities) {
                        $scheduleSubscribersOffset = $this->configService->get('June.config.subscriberOffset');
                        $this->loggerHelper->logInfo("Scheduled subscribers export started for " . $scheduleSubscribersOffset . "-" . ($scheduleSubscribersOffset + self::SUBSCRIBER_LIMIT_PER_REQUEST), "exportSubscribers");
                        $this->configService->set('June.config.subscriberOffset', ($scheduleSubscribersOffset + self::SUBSCRIBER_LIMIT_PER_REQUEST));
                    } else {
                        $scheduleCustomersOffset = $this->configService->get('June.config.customerOffset');
                        $this->loggerHelper->logInfo("Scheduled customers export started for " . $scheduleCustomersOffset . "-" . ($scheduleCustomersOffset + self::SUBSCRIBER_LIMIT_PER_REQUEST), "exportSubscribers");
                        $this->configService->set('June.config.customerOffset', ($scheduleCustomersOffset + self::SUBSCRIBER_LIMIT_PER_REQUEST));
                    }
                }
            } catch (\Exception $e) {
                $this->loggerHelper->logError($e->getMessage()." - ".$e->getTraceAsString(), "exportSubscribers");
            }
        } else {
            if (empty($commerceCollectToken)) {
                $this->loggerHelper->logInfo("Scheduled subscribers export could not run because CommerceCollectToken is not configured in the JUNE extension configuration.", "exportSubscribers");
            } elseif (empty($commerceCollectAuth)) {
                $this->loggerHelper->logInfo("Scheduled subscribers export could not run because CommerceCollectAUTH is not configured in the JUNE extension configuration.", "exportSubscribers");
            }
        }
    }

    /**
     * Get new set of newsletter subscribers to be exported
     * @return array<NewsletterRecipientEntity>
     */
    private function getNewsletterSubscribers(): array {
        $pluginCriteria = new Criteria();
        $pluginCriteria = $pluginCriteria->addFilter(new EqualsFilter('name', 'June'));
        $plugin = $this->pluginRepository->search($pluginCriteria, Context::createDefaultContext())->first();
        $installDate = $plugin->getInstalledAt()->format('c');

        $offset = $this->configService->get('June.config.subscriberOffset') ?? 0;

        $subscriberCriteria = new Criteria();
        $subscriberCriteria->addFilter(new RangeFilter('createdAt',
            [
                RangeFilter::LTE => $installDate
            ]
        ));
        $subscriberCriteria->addAssociation('tags');
        $subscriberCriteria->addAssociation('salutation');
        $subscriberCriteria->setLimit(self::SUBSCRIBER_LIMIT_PER_REQUEST);
        $subscriberCriteria->setOffset($offset);

        return $this->newsletterSubscriberRepository->search($subscriberCriteria, Context::createDefaultContext())->getElements();
    }

    /**
     * Get total count of newsletter subscribers to be exported
     * @return int
     */
    private function getTotalNewsletterSubscribersCount(): int {
        $pluginCriteria = new Criteria();
        $pluginCriteria = $pluginCriteria->addFilter(new EqualsFilter('name', 'June'));
        $plugin = $this->pluginRepository->search($pluginCriteria, Context::createDefaultContext())->first();
        $installDate = $plugin->getInstalledAt()->format('c');

        $subscriberCriteria = new Criteria();
        $subscriberCriteria->addFilter(new RangeFilter('createdAt',
            [
                RangeFilter::LTE => $installDate
            ]
        ));

        return $this->newsletterSubscriberRepository->search($subscriberCriteria, Context::createDefaultContext())->count();
    }

    /**
     * Get new set of customers to be exported
     * @return array<CustomerEntity>
     */
    private function getCustomers(): array {
        $pluginCriteria = new Criteria();
        $pluginCriteria = $pluginCriteria->addFilter(new EqualsFilter('name', 'June'));
        $plugin = $this->pluginRepository->search($pluginCriteria, Context::createDefaultContext())->first();
        $installDate = $plugin->getInstalledAt()->format('c');

        $offset = $this->configService->get('June.config.customerOffset') ?? 0;

        $customerCriteria = new Criteria();
        $customerCriteria->addFilter(new RangeFilter('createdAt',
            [
                RangeFilter::LTE => $installDate
            ]
        ));
        $customerCriteria->addAssociation('tags');
        $customerCriteria->addAssociation('salutation');
        $customerCriteria->setLimit(self::SUBSCRIBER_LIMIT_PER_REQUEST);
        $customerCriteria->setOffset($offset);

        return $this->customerRepository->search($customerCriteria, Context::createDefaultContext())->getElements();
    }

    /**
     * Get total count of customers to be exported
     * @return int
     */
    private function getTotalCustomersCount(): int {
        $pluginCriteria = new Criteria();
        $pluginCriteria = $pluginCriteria->addFilter(new EqualsFilter('name', 'June'));
        $plugin = $this->pluginRepository->search($pluginCriteria, Context::createDefaultContext())->first();
        $installDate = $plugin->getInstalledAt()->format('c');

        $customerCriteria = new Criteria();
        $customerCriteria->addFilter(new RangeFilter('createdAt',
            [
                RangeFilter::LTE => $installDate
            ]
        ));

        return $this->customerRepository->search($customerCriteria, Context::createDefaultContext())->count();
    }

    /**
     * Get single customer entity by Id
     * @param NewsletterRecipientEntity $subscriber
     * @return CustomerEntity|null
     */
    private function getCustomer(NewsletterRecipientEntity $subscriber): ?CustomerEntity {
        if ($subscriber->getEmail()) {
            $itemCriteria = new Criteria();
            $itemCriteria->addFilter(new EqualsFilter("email", $subscriber->getEmail()));
            $itemCriteria->addAssociation('tags');
            return $this->customerRepository->search($itemCriteria, Context::createDefaultContext())->first();
        }
        return null;
    }

    /**
     * @param CustomerEntity $customer
     * @return NewsletterRecipientEntity|null
     */
    private function getNewsletterSubscriber(CustomerEntity $customer): ?NewsletterRecipientEntity {
        if ($customer->getEmail()) {
            $itemCriteria = new Criteria();
            $itemCriteria->addFilter(new EqualsFilter("email", $customer->getEmail()));
            $itemCriteria->addAssociation('tags');
            return $this->newsletterSubscriberRepository->search($itemCriteria, Context::createDefaultContext())->first();
        }
        return null;
    }

    /**
     * @param NewsletterRecipientEntity|CustomerEntity $user
     * @return LocaleEntity|null
     */
    private function getLocale(NewsletterRecipientEntity|CustomerEntity $user): ?LocaleEntity {
        $locale = null;

        if ($user && $user->getLanguageId()) {
            $languageCriteria = new Criteria();
            $languageCriteria = $languageCriteria->addFilter(new EqualsFilter('id', $user->getLanguageId()));
            $language = $this->languageRepository->search($languageCriteria, Context::createDefaultContext())->first();

            if ($language->getLocaleId()) {
                $localeCriteria = new Criteria();
                $localeCriteria = $localeCriteria->addFilter(new EqualsFilter('id', $language->getLocaleId()));
                $locale = $this->localeRepository->search($localeCriteria, Context::createDefaultContext())->first();
            }
        }

        return $locale;
    }

    /**
     * @param NewsletterRecipientEntity|CustomerEntity $user
     * @return CountryEntity|null
     */
    private function getCountry(NewsletterRecipientEntity|CustomerEntity $user): ?CountryEntity {
        $country = null;

        if ($user && $user->getSalesChannelId()) {
            $salesChannelCriteria = new Criteria();
            $salesChannelCriteria = $salesChannelCriteria->addFilter(new EqualsFilter('id', $user->getSalesChannelId()));
            $salesChannel = $this->salesChannelRepository->search($salesChannelCriteria, Context::createDefaultContext())->first();

            if ($salesChannel->getCountryId()) {
                $countryCriteria = new Criteria();
                $countryCriteria = $countryCriteria->addFilter(new EqualsFilter('id', $salesChannel->getCountryId()));
                $country = $this->countryRepository->search($countryCriteria, Context::createDefaultContext())->first();
            }
        }

        return $country;
    }

    /**
     * @param string $tagId
     * @return string
     */
    private function getTag(string $tagId): string {
        if (!in_array($tagId, $this->loadedSubscriberTags)) {
            $tag = $this->customerTagsRepository->search(new Criteria([$tagId]), Context::createDefaultContext())->first();
            $this->loadedSubscriberTags[$tagId] = $tag->getName();
        }

        return $this->loadedSubscriberTags[$tagId];
    }
}
