<?php declare(strict_types=1);

namespace June\Service\ScheduledTask;

use June\Components\Api\Client\JuneApiClient;
use June\Components\Logger\Helper\LoggerHelper;
use Shopware\Core\Checkout\Customer\CustomerEntity;
use Shopware\Core\Framework\Context;
use Shopware\Core\Framework\DataAbstractionLayer\EntityRepository;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Criteria;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Filter\EqualsFilter;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Filter\RangeFilter;
use Shopware\Core\Framework\MessageQueue\ScheduledTask\ScheduledTask;
use Shopware\Core\Framework\MessageQueue\ScheduledTask\ScheduledTaskDefinition;
use Shopware\Core\Framework\MessageQueue\ScheduledTask\ScheduledTaskEntity;
use Shopware\Core\Framework\MessageQueue\ScheduledTask\ScheduledTaskHandler;
use Shopware\Core\System\SystemConfig\SystemConfigService;
use Shopware\Core\Content\Newsletter\Aggregate\NewsletterRecipient\NewsletterRecipientEntity;
use June\Subscriber\JuneCommerceSubscriber;

use Shopware\Core\System\Locale\LocaleEntity;
use Shopware\Core\System\Country\CountryEntity;
use Shopware\Core\System\Tag\TagEntity;

#[AsMessageHandler(handles: ExportSubscriberUpdatesTask::class)]
class ExportSubscriberUpdatesTaskHandler extends ScheduledTaskHandler {

    const SUBSCRIBER_LIMIT_PER_REQUEST = 200;
    const CONF_CURSOR_DATE_POSITION = 'June.config.last_subscriber_updated_at_cursor';
    const CONF_CURSOR_PAGINATION_POSITION = 'June.config.last_subscriber_pagination_cursor';

    private EntityRepository $languageRepository;

    private EntityRepository $localeRepository;

    private EntityRepository $countryRepository;

    private EntityRepository $newsletterSubscriberRepository;

    private EntityRepository $customerRepository;

    private EntityRepository $customerTagsRepository;

    private EntityRepository $salesChannelRepository;

    private SystemConfigService $configService;

    private LoggerHelper $loggerHelper;

    private JuneApiClient $juneApiClient;

    private string $lastExecutionTime;

    private array $loadedSubscriberTags = [];

    public function __construct
    (
        EntityRepository $scheduledTaskRepository,
        EntityRepository $languageRepository,
        EntityRepository $localeRepository,
        EntityRepository $countryRepository,
        EntityRepository $newsletterSubscriberRepository,
        EntityRepository $customerRepository,
        EntityRepository $customerTagsRepository,
        EntityRepository $salesChannelRepository,
        SystemConfigService $configService,
        LoggerHelper $loggerHelper,
        JuneApiClient $juneApiClient
    )
    {
        parent::__construct($scheduledTaskRepository);
        $this->languageRepository = $languageRepository;
        $this->localeRepository = $localeRepository;
        $this->countryRepository = $countryRepository;
        $this->newsletterSubscriberRepository = $newsletterSubscriberRepository;
        $this->customerRepository = $customerRepository;
        $this->customerTagsRepository = $customerTagsRepository;
        $this->salesChannelRepository = $salesChannelRepository;
        $this->configService = $configService;
        $this->loggerHelper = $loggerHelper;
        $this->juneApiClient = $juneApiClient;
    }

    /**
     * @param ScheduledTask $task
     * @return void
     */
    public function __invoke(ScheduledTask $task): void
    {
        $taskId = $task->getTaskId();
        if ($taskId === null) {
            // run task independent of the schedule
            $this->run();

            return;
        }

        /** @var ScheduledTaskEntity|null $taskEntity */
        $taskEntity = $this->scheduledTaskRepository
            ->search(new Criteria([$taskId]), Context::createDefaultContext())
            ->get($taskId);

        if ($taskEntity === null || !$taskEntity->isExecutionAllowed()) {
            return;
        }

        try {
            if ($this->configService->get(self::CONF_CURSOR_DATE_POSITION)) {
                $this->lastExecutionTime = $this->configService->get(self::CONF_CURSOR_DATE_POSITION);
            } else if ($taskEntity->getLastExecutionTime()) {
                $this->lastExecutionTime = $taskEntity->getLastExecutionTime()->format('c');
            }
            //$this->loggerHelper->logInfo("starting exportSubscriberUpdates. Last execution: ".$this->lastExecutionTime, "exportSubscriberUpdates");
            $this->markTaskRunning($task);
            $this->run();
        } catch (\Throwable $e) {
            $this->markTaskFailed($task);

            $this->loggerHelper->logError($e->getMessage()." - ".$e->getTraceAsString(), "exportSubscriberUpdates");
        }

        $this->rescheduleTask($task, $taskEntity);
    }

    /**
     * @return iterable
     */
    public static function getHandledMessages(): iterable
    {
        return [ExportSubscriberUpdatesTask::class];
    }

    /**
     * @return void
     */
    public function run(): void
    {
        $commerceCollectToken = $this->configService->get(JuneCommerceSubscriber::JUNE_COMMERCE_COLLECT_TOKEN);
        $commerceCollectAuth = $this->configService->get(JuneCommerceSubscriber::JUNE_COMMERCE_COLLECT_AUTH);

        if($this->configService->get(JuneCommerceSubscriber::JUNE_TRANSFER_CUSTOMERS) && !empty($commerceCollectToken) && !empty($commerceCollectAuth)) {
            try {
                $customerData = [];
                $customerUpdateFields = [
                    "external_id",
                    "firstname",
                    "lastname",
                    "salutation",
                    "salutation_readable",
                    "locale",
                    "updated_at",
                    "guest"
                ];

                $subscribers = $this->getNewsletterSubscribers();
                if ($subscribers) {
                    $limitReached = count($subscribers)>=self::SUBSCRIBER_LIMIT_PER_REQUEST;
                    $lastUpdatedDate = null;
                    foreach ($subscribers as $subscriber) {
                        if (!$subscriber || !$subscriber->getId() || !$subscriber->getEmail()) {
                            continue;
                        }
                        $lastUpdatedDate = $subscriber->getUpdatedAt();

                        $customer = $this->getCustomer($subscriber);
                        $locale = $this->getLocale($subscriber);
                        $country = $this->getCountry($subscriber);

                        $enrichedCustomer = [
                            'external_id'               => $customer?->getCustomerNumber() ?? $subscriber?->getId(),
                            'email'                     => $subscriber->getEmail(),
                            'firstname'                 => $customer?->getFirstname() ?? $subscriber->getFirstname(),
                            'lastname'                  => $customer?->getLastname() ?? $subscriber->getLastname(),
                            'salutation'                => $customer?->getSalutation()?->getSalutationKey() ?? $subscriber->getSalutation()?->getSalutationKey(),
                            'salutation_readable'       => $customer?->getSalutation()?->getLetterName() ?? $subscriber->getSalutation()?->getLetterName(),
                            'store_id'                  => $subscriber->getSalesChannelId() ?? $customer?->getSalesChannelId(),
                            'locale'                    => $locale?->getCode(),
                            'created_at'                => $customer?->getCreatedAt()?->format('c') ?? $subscriber->getCreatedAt()?->format('c'),
                            'updated_at'                => $customer?->getUpdatedAt()?->format('c') ?? $subscriber->getUpdatedAt()?->format('c'),
                            'geo_location'              => [
                                'country'                   => $country?->getName(),
                            ],
                            'guest'                     => !$customer || $customer->getGuest()
                        ];

                        // add more information to customer if available
                        if ($subscriber?->getConfirmedAt()) {
                            $enrichedCustomer['subscription_confirmed_at'] = $subscriber->getConfirmedAt()->format('c');
                            $customerUpdateFields[] = 'subscription_confirmed_at';
                        }

                        if ($customer?->getCustomFields()) {
                            foreach ($customer->getCustomFields() as $fieldName => $fieldValue) {
                                if (!isset($enrichedCustomer[$fieldName])) {
                                    $transformedFieldName = ($enrichedCustomer['guest'] ? 'guest_' : '') . $fieldName;
                                    $enrichedCustomer[$transformedFieldName] = $fieldValue;

                                    if (!in_array($transformedFieldName, $customerUpdateFields)) {
                                        $customerUpdateFields[] = $transformedFieldName;
                                    }
                                }
                            }
                        }
                        if ($customer?->getBirthday()) {
                            $enrichedCustomer['birthday'] = $customer->getBirthday()->format('c');
                            if (!in_array('birthday', $customerUpdateFields)) {
                                $customerUpdateFields[] = 'birthday';
                            }
                        }
                        if ($customer?->getLastLogin()) {
                            $enrichedCustomer['last_login'] = $customer->getLastLogin()->format('c');
                            if (!in_array('last_login', $customerUpdateFields)) {
                                $customerUpdateFields[] = 'last_login';
                            }
                        }
                        if ($customer?->getLastOrderDate()) {
                            $enrichedCustomer['last_order_date'] = $customer->getLastOrderDate()->format('c');
                            if (!in_array('last_order_date', $customerUpdateFields)) {
                                $customerUpdateFields[] = 'last_order_date';
                            }
                        }
                        if ($customer?->getOrderCount()) {
                            $enrichedCustomer['order_count'] = $customer->getOrderCount();
                            if (!in_array('order_count', $customerUpdateFields)) {
                                $customerUpdateFields[] = 'order_count';
                            }
                        }
                        if ($customer?->getOrderTotalAmount()) {
                            $enrichedCustomer['order_total_amount'] = $customer->getOrderTotalAmount();
                            if (!in_array('order_total_amount', $customerUpdateFields)) {
                                $customerUpdateFields[] = 'order_total_amount';
                            }
                        }

                        if ($customer?->getCampaignCode()) {
                            $enrichedCustomer['campaign_code'] = $customer->getCampaignCode();
                            if (!in_array('campaign_code', $customerUpdateFields)) {
                                $customerUpdateFields[] = 'campaign_code';
                            }
                        }

                        if ($customer?->getAffiliateCode()) {
                            $enrichedCustomer['affiliate_code'] = $customer->getAffiliateCode();
                            if (!in_array('affiliate_code', $customerUpdateFields)) {
                                $customerUpdateFields[] = 'affiliate_code';
                            }
                        }

                        $enrichedCustomer['_unsubscribed'] = !($subscriber?->getStatus() === "direct" || $subscriber?->getStatus() === "optIn");
                        if (!in_array('_unsubscribed', $customerUpdateFields)) {
                            $customerUpdateFields[] = '_unsubscribed';
                        }

                        $enrichedCustomer['intent'] = $subscriber?->getStatus();
                        if (!in_array('intent', $customerUpdateFields)) {
                            $customerUpdateFields[] = 'intent';
                        }

                        if ($subscriber?->getCity()) {
                            $enrichedCustomer['city'] = $subscriber->getCity();
                            $enrichedCustomer['geo_location']['city'] = $subscriber->getCity();
                            if (!in_array('city', $customerUpdateFields)) {
                                $customerUpdateFields[] = 'city';
                            }
                        }
                        if ($subscriber?->getZipCode()) {
                            $enrichedCustomer['zip_code'] = $subscriber->getZipCode();
                            if (!in_array('zip_code', $customerUpdateFields)) {
                                $customerUpdateFields[] = 'zip_code';
                            }
                        }

                        if ($customer?->getTagIds()) {
                            $enrichedCustomer['_tags'] = [];
                            foreach($customer->getTagIds() AS $tagId) {
                                if ($tag = $this->getTag($tagId)) {
                                    $enrichedCustomer['_tags'][] = $tag;
                                }
                            }
                            if (!in_array('_tags', $customerUpdateFields)) {
                                $customerUpdateFields[] = '_tags';
                            }
                        }
                        if ($subscriber?->getTags()) {
                            if (!isset($enrichedCustomer['_tags'])) {
                                $enrichedCustomer['_tags'] = [];
                            }
                            foreach ($subscriber->getTags() as $tag) {
                                if ($tag?->getName() && !in_array($tag->getName(), $enrichedCustomer['_tags'])) {
                                    $enrichedCustomer['_tags'][] = $tag->getName();
                                }
                            }
                            if (!in_array('_tags', $customerUpdateFields)) {
                                $customerUpdateFields[] = '_tags';
                            }
                        }

                        $customerData[$subscriber->getEmail()] = $enrichedCustomer;
                    }

                    if ($limitReached) {
                        // save updatedAtDate for next execution
                        $this->configService->set(self::CONF_CURSOR_PAGINATION_POSITION, $this->configService->get(self::CONF_CURSOR_PAGINATION_POSITION) + 1);
                        $this->configService->set(self::CONF_CURSOR_DATE_POSITION, $lastUpdatedDate->format('c'));
                    } else {
                        // remove updatedAtDate for next execution (then uses last scheduleExecutionDate)
                        $this->configService->set(self::CONF_CURSOR_PAGINATION_POSITION, 0);
                        $this->configService->set(self::CONF_CURSOR_DATE_POSITION, null);
                    }

                    // send collected data to June
                    if (!empty($customerData)) {
                        $this->loggerHelper->logInfo("Scheduled subscribers update export started for ".count($subscribers)." subscribers.", "exportSubscriberUpdates");

                        $requestData = [
                            "action"            => "subscriber_update",
                            "json"              => array_values($customerData),
                            "activate_trigger"  => true,
                            "default_subscribe" => false,
                            "force_subscribe"   => false,
                            "update_if_exists"  => true,
                            "overwrite_tags"    => true,
                            "update_fields"     => $customerUpdateFields,
                            "last_updated_at"   => $this->configService->get(self::CONF_CURSOR_DATE_POSITION),
                        ];

                        $this->juneApiClient->sendData($requestData, "https://engagement.juneapp.com/collect/$commerceCollectToken/import/json", $commerceCollectAuth, 60);
                    }
                }
            } catch (\Exception $e) {
                $this->loggerHelper->logError($e->getMessage()." - ".$e->getTraceAsString(), "exportSubscriberUpdates");
            }
        } else {
            if (empty($commerceCollectToken)) {
                $this->loggerHelper->logInfo("Scheduled subscribers update export could not run because CommerceCollectToken is not configured in the JUNE extension configuration.", "exportSubscriberUpdates");
            } elseif (empty($commerceCollectAuth)) {
                $this->loggerHelper->logInfo("Scheduled subscribers update export could not run because CommerceCollectAUTH is not configured in the JUNE extension configuration.", "exportSubscriberUpdates");
            }
        }
    }

    /**
     * @return array<NewsletterRecipientEntity>
     */
    private function getNewsletterSubscribers(): array {
        if (!$this->lastExecutionTime) {
            $now = new \DateTime();
            $minutes10Ago = new \DateTime();
            $minutes10Ago->setTimestamp($now->getTimestamp()-(60*10));
            $this->lastExecutionTime = $minutes10Ago->format('c');
        }

        $lastUpdatedSubscriberPage = $this->configService->get(self::CONF_CURSOR_PAGINATION_POSITION) ?? 0;

        // format updatedAt date and remove offset because Shopware adds it themselves
        $formatedUtcNoOffset = new \DateTime($this->lastExecutionTime);
        $formatedUtcNoOffset = $formatedUtcNoOffset->format('Y-m-d\TH:i:s');

        $subscriberCriteria = new Criteria();
        $subscriberCriteria->addFilter(new RangeFilter('updatedAt',
            [
                RangeFilter::GTE => $formatedUtcNoOffset
            ]
        ));
        if ($lastUpdatedSubscriberPage) {
            $subscriberCriteria->setOffset($lastUpdatedSubscriberPage * self::SUBSCRIBER_LIMIT_PER_REQUEST);
        }
        $subscriberCriteria->setLimit(self::SUBSCRIBER_LIMIT_PER_REQUEST);
        $subscriberCriteria->addAssociation('tags');
        $subscriberCriteria->addAssociation('salutation');

        // $this->loggerHelper->logInfo("Scheduled subscribers query: Offset:".$lastUpdatedSubscriberPage * self::SUBSCRIBER_LIMIT_PER_REQUEST ." - Limit:".self::SUBSCRIBER_LIMIT_PER_REQUEST ." - UpdatedAt>=".$formatedUtcNoOffset, "exportSubscriberUpdates");

        return $this->newsletterSubscriberRepository->search($subscriberCriteria, Context::createDefaultContext())->getElements();
    }

    /**
     * @param NewsletterRecipientEntity $subscriber
     * @return CustomerEntity|null
     */
    private function getCustomer(NewsletterRecipientEntity $subscriber): ?CustomerEntity {
        if ($subscriber && $subscriber->getEmail()) {
            $itemCriteria = new Criteria();
            $itemCriteria->addFilter(new EqualsFilter("email", $subscriber->getEmail()));
            $itemCriteria->addAssociation('tags');
            return $this->customerRepository->search($itemCriteria, Context::createDefaultContext())->first();
        }
        return null;
    }

    /**
     * @param NewsletterRecipientEntity $subscriber
     * @return LocaleEntity|null
     */
    private function getLocale(NewsletterRecipientEntity $subscriber): ?LocaleEntity {
        $locale = null;

        if ($subscriber->getLanguageId()) {
            $languageCriteria = new Criteria();
            $languageCriteria = $languageCriteria->addFilter(new EqualsFilter('id', $subscriber->getLanguageId()));
            $language = $this->languageRepository->search($languageCriteria, Context::createDefaultContext())->first();

            if ($language->getLocaleId()) {
                $localeCriteria = new Criteria();
                $localeCriteria = $localeCriteria->addFilter(new EqualsFilter('id', $language->getLocaleId()));
                $locale = $this->localeRepository->search($localeCriteria, Context::createDefaultContext())->first();
            }
        }

        return $locale;
    }

    /**
     * @param NewsletterRecipientEntity $subscriber
     * @return CountryEntity|null
     */
    private function getCountry(NewsletterRecipientEntity $subscriber): ?CountryEntity {
        $country = null;

        if ($subscriber->getSalesChannelId()) {
            $salesChannelCriteria = new Criteria();
            $salesChannelCriteria = $salesChannelCriteria->addFilter(new EqualsFilter('id', $subscriber->getSalesChannelId()));
            $salesChannel = $this->salesChannelRepository->search($salesChannelCriteria, Context::createDefaultContext())->first();

            if ($salesChannel->getCountryId()) {
                $countryCriteria = new Criteria();
                $countryCriteria = $countryCriteria->addFilter(new EqualsFilter('id', $salesChannel->getCountryId()));
                $country = $this->countryRepository->search($countryCriteria, Context::createDefaultContext())->first();
            }
        }

        return $country;
    }

    /**
     * @param string $tagId
     * @return string
     */
    private function getTag(string $tagId): string {
        if (!in_array($tagId, $this->loadedSubscriberTags)) {
            $tag = $this->customerTagsRepository->search(new Criteria([$tagId]), Context::createDefaultContext())->first();
            $this->loadedSubscriberTags[$tagId] = $tag->getName();
        }

        return $this->loadedSubscriberTags[$tagId];
    }
}
