<?php

namespace June\Subscriber;

use June\Components\Api\Client\JuneApiClient;

use Shopware\Core\Checkout\Customer\CustomerEntity;
use Shopware\Core\Checkout\Order\Aggregate\OrderCustomer\OrderCustomerEntity;
use Shopware\Core\Content\Newsletter\Aggregate\NewsletterRecipient\NewsletterRecipientEntity;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Shopware\Core\Checkout\Order\OrderEvents;
use Shopware\Core\Checkout\Cart\Event\CheckoutOrderPlacedEvent;
use Shopware\Core\Checkout\Cart\Event\CartChangedEvent;
use Shopware\Core\Checkout\Customer\Event\GuestCustomerRegisterEvent;
use Shopware\Core\Checkout\Customer\Event\CustomerRegisterEvent;

use Shopware\Core\Content\Newsletter\NewsletterEvents;
use Shopware\Core\Content\Newsletter\Event\NewsletterRegisterEvent;
use Shopware\Core\Content\Newsletter\Event\NewsletterConfirmEvent;
use Shopware\Core\Content\Newsletter\Event\NewsletterUnsubscribeEvent;
use Shopware\Core\Framework\DataAbstractionLayer\Event\EntityDeletedEvent;

use Shopware\Core\Checkout\Cart\LineItem\LineItem;
use Shopware\Core\Checkout\Order\OrderEntity;
use Shopware\Core\Checkout\Order\Aggregate\OrderLineItem\OrderLineItemEntity;

use Shopware\Core\Framework\DataAbstractionLayer\Event\EntityWrittenEvent;
use Shopware\Core\System\SalesChannel\Context\AbstractSalesChannelContextFactory;
use Shopware\Core\System\SalesChannel\Entity\SalesChannelRepository;
use Shopware\Core\System\SalesChannel\SalesChannelContext;
use Shopware\Core\System\SystemConfig\SystemConfigService;
use Shopware\Core\Framework\DataAbstractionLayer\EntityRepository;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Filter\EqualsFilter;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Criteria;
use Shopware\Core\Framework\Context;
use June\Components\Logger\Helper\LoggerHelper;

class JuneCommerceSubscriber implements EventSubscriberInterface
{
    const JUNE_COMMERCE_TOKEN = "June.config.commercetoken";
    const JUNE_COMMERCE_COLLECT_TOKEN = "June.config.commercecollecttoken";
    const JUNE_COMMERCE_COLLECT_AUTH = "June.config.commercecollectauth";
    const JUNE_TRANSFER_ORDER = "June.config.orderCommerceData";
    const JUNE_TRANSFER_FULFILLMENT = "June.config.fulfillCommerceData";
    const JUNE_TRANSFER_CART = "June.config.cartCommerceData";
    const JUNE_TRANSFER_GUESTDATA = "June.config.transferGuestData";
    const JUNE_TRANSFER_CUSTOMERS = "June.config.transferCustomers";
    const JUNE_TRANSFER_SUBSCRIBERS = "June.config.transferSubscribers";

    private EntityRepository $salesRepository;

    private EntityRepository $orderRepository;

    private EntityRepository $orderLineItemsRepository;

    private EntityRepository $languageRepository;

    private EntityRepository $localeRepository;

    private EntityRepository $countryRepository;

    private EntityRepository $customerEntity;

    private EntityRepository $newsletterRecipientRepository;

    private EntityRepository $customerTagsRepository;

    private SalesChannelRepository $categoryRepository;

    private EntityRepository $salutationRepository;

    private AbstractSalesChannelContextFactory $salesChannelContextFactory;

    private SystemConfigService $configService;

    private LoggerHelper $loggerHelper;

    private JuneApiClient $juneApiClient;

    private array $loadedCustomerTags = [];

    public function __construct
    (
        EntityRepository $salesRepository,
        EntityRepository $orderRepository,
        EntityRepository $orderLineItemsRepository,
        EntityRepository $languageRepository,
        EntityRepository $localeRepository,
        EntityRepository $countryRepository,
        EntityRepository $customerEntity,
        EntityRepository $newsletterRecipientRepository,
        EntityRepository $customerTagsRepository,
        EntityRepository $salutationRepository,
        SalesChannelRepository $categoryRepository,
        AbstractSalesChannelContextFactory $salesChannelContextFactory,
        SystemConfigService $configService,
        LoggerHelper $loggerHelper,
        JuneApiClient $juneApiClient
    )
    {
        $this->salesRepository = $salesRepository;
        $this->orderRepository = $orderRepository;
        $this->orderLineItemsRepository = $orderLineItemsRepository;
        $this->languageRepository = $languageRepository;
        $this->localeRepository = $localeRepository;
        $this->countryRepository = $countryRepository;
        $this->customerEntity = $customerEntity;
        $this->newsletterRecipientRepository = $newsletterRecipientRepository;
        $this->customerTagsRepository = $customerTagsRepository;
        $this->salutationRepository = $salutationRepository;
        $this->categoryRepository = $categoryRepository;
        $this->salesChannelContextFactory = $salesChannelContextFactory;
        $this->configService = $configService;
        $this->loggerHelper = $loggerHelper;
        $this->juneApiClient = $juneApiClient;
    }

    public static function getSubscribedEvents(): array
    {
        return [
            CartChangedEvent::class => 'cartChanged',
            // CheckoutOrderPlacedEvent::EVENT_NAME => 'orderPlaced', // EVENT_NAME does not work? What is Shopware doing?
            CheckoutOrderPlacedEvent::class => 'orderPlaced',
            OrderEvents::ORDER_DELIVERY_WRITTEN_EVENT => 'orderFulfilled',
            GuestCustomerRegisterEvent::EVENT_NAME => 'guestRegister',
            CustomerRegisterEvent::EVENT_NAME => 'customerRegister',
            NewsletterEvents::NEWSLETTER_REGISTER_EVENT => 'newsletterRegister',
            NewsletterEvents::NEWSLETTER_CONFIRM_EVENT => 'newsletterSubscribe',
            NewsletterEvents::NEWSLETTER_UNSUBSCRIBE_EVENT => 'newsletterUnsubscribe',
            NewsletterEvents::NEWSLETTER_RECIPIENT_DELETED_EVENT => 'newsletterUnsubscribe',

            # generic events for all entities
            //EntityWriteEvent::class => 'genericEntityWriteEvent',
        ];
    }

    /**
     * @param CartChangedEvent $event
     * @return void
     */
    public function cartChanged(CartChangedEvent $event) {
        $juneCommerceToken = $this->configService->get(self::JUNE_COMMERCE_TOKEN);
        if(!empty($juneCommerceToken)) {
            if($this->configService->get(self::JUNE_TRANSFER_CART)) {
                // make sure it's a registered customer or we are supposed to transfer guests
                if(
                    $this->configService->get(self::JUNE_TRANSFER_GUESTDATA) ||
                    !$event->getContext()?->getCustomer()?->getGuest()
                ) {
                    try {
                        $customer = $event->getContext()?->getCustomer();

                        // make sure we have enough customer information to use it
                        if ($customer?->getEmail()) {
                            $shopUrl = $this->getShopUrl($event->getContext()?->getSalesChannelId(), $event->getContext()?->getLanguageId());
                            $products = [];
                            $data = [];

                            if ($customer) {
                                $data['id'] = ($customer->getCustomerNumber() ?? $customer->getId());
                                $data['email'] = $customer->getEmail();
                                $data['customer'] = [
                                    'firstname' => $customer->getFirstName(),
                                    'lastname' => $customer->getLastName(),
                                    'store_id' => $customer->getSalesChannelId(),
                                    'locale' => $this->getLanguageCode($event->getContext()?->getCustomer()?->getLanguageId()),
                                    'geo_location' => [
                                        'city' => ($customer->getDefaultBillingAddress()?->getCity()),
                                        'continent' => '',
                                        'country' => $this->getCountry($event->getContext()->getSalesChannelId()),
                                    ],
                                    'guest' => $customer->getGuest(),
                                ];
                            } else {
                                $data['id'] = session_id();
                            }

                            foreach ($event->getCart()->getLineItems() as $item) {
                                if ($product = $this->transformLineItem($item, $shopUrl, $event->getContext()->getSalesChannelId(), $voucher)) {
                                    $products[] = $product;
                                }
                            }

                            $data['cart'] = [
                                'action' => 'update',
                                'products' => $products
                            ];

                            $this->juneApiClient->sendData($data, "https://engagement.juneapp.com/commerce/webhook/$juneCommerceToken", null);
                        }
                    } catch (\Exception $e) {
                        $this->loggerHelper->logError($e->getMessage(), "commerceCartChanged");
                    }
                }
            }
        } else {
            $this->loggerHelper->logError("The Commerce Token in the JUNE Plugin configuration is missing", "pluginConfig");
        }
    }

    /**
     * @param CheckoutOrderPlacedEvent $event
     * @return void
     */
    public function orderPlaced(CheckoutOrderPlacedEvent $event) {
        $juneCommerceToken = $this->configService->get(self::JUNE_COMMERCE_TOKEN);
        if(!empty($juneCommerceToken)) {
            if($this->configService->get(self::JUNE_TRANSFER_ORDER)) {
                if(
                    $this->configService->get(self::JUNE_TRANSFER_GUESTDATA) ||
                    !$event->getOrder()->getOrderCustomer()?->getCustomer()?->getGuest()
                ) {
                    try {
                        $order = $event->getOrder();

                        $voucher = 0;
                        $shopUrl = $this->getShopUrl($order->getSalesChannelId(), $order->getOrderCustomer()?->getCustomer()?->getLanguageId());

                        $recipient = array_keys($event->getMailStruct()->getRecipients());
                        $orderDate = $order->getOrderDateTime()->format('c');
                        $customer = $order->getOrderCustomer()->getCustomer() ?? $order->getOrderCustomer();
                        $products = [];

                        $city = "";
                        $billingId = $order->getBillingAddressId();
                        foreach ($order->getAddresses() as $address) {
                            if($address->getId() == $billingId) {
                                $city = $address->getCity();
                            }
                        }

                        foreach ($order->getLineItems() as $item) {
                            // transform products and add voucher values to referenced variable
                            if ($product = $this->transformLineItem($item, $shopUrl, $order->getSalesChannelId(), $voucher)) {
                                $products[] = $product;
                            }
                        }

                        $data = [
                            'action'    => 'order_placed',
                            'id'        => ($customer ? ($customer->getCustomerNumber() ?? $customer->getId()) : session_id()),
                            'email'     => $recipient[0],
                            'customer'  => [
                                'firstname'     => $customer?->getFirstname(),
                                'lastname'      => $customer?->getLastname(),
                                'store_id'      => $customer?->getSalesChannelId(),
                                'locale'        => $this->getLanguageCode($customer->getLanguageId()),
                                'geo_location'  => [
                                    'city'          => $city,
                                    'country'       => $this->getCountry($order->getSalesChannelId()),
                                ],
                                'guest'         => $customer?->getGuest()
                            ],
                            'order'     => [
                                'id'            => $order->getOrderNumber(),
                                'created_at'    => $orderDate,
                                'fulfilled_at'  => null,
                                'total'         => $order->getPrice()?->getTotalPrice(),
                                'shipping'      => $order->getShippingCosts()?->getTotalPrice(),
                                'voucher'       => $voucher,
                                'products'      => $products,
                            ],
                        ];

                        // set new updatedAt date for customer so all properties are exported on the next schedule run
                        if ($this->configService->get(self::JUNE_TRANSFER_CUSTOMERS)) {
                            $this->createJuneCustomer($customer, $customer?->getSalesChannelId() ?? $order->getSalesChannelId(), $this->getLanguageCode($customer->getLanguageId()));

                            $customerId = ($customer instanceof CustomerEntity ? $customer->getId() : $customer->getCustomerId());
                            if ($customerId) {
                                $this->customerEntity->update([
                                    [
                                        'id' => $customerId
                                    ]
                                ], $event->getContext());
                            }
                        }

                        // send order information
                        $this->juneApiClient->sendData($data,"https://engagement.juneapp.com/commerce/webhook/$juneCommerceToken", null);
                    } catch(\Exception $e) {
                        $this->loggerHelper->logError($e->getMessage()." - ".$e->getTraceAsString(), "commerceOrderPlaced");
                    }
                }
            }
        } else {
            $this->loggerHelper->logError("The Commerce Token in the JUNE Plugin configuration is missing", "pluginConfig");
        }
    }

    /**
     * @param EntityWrittenEvent $event
     * @return void
     */
    public function orderFulfilled(EntityWrittenEvent $event): void {
        $juneCommerceToken = $this->configService->get(self::JUNE_COMMERCE_TOKEN);
        if(!empty($juneCommerceToken)) {
            if ($this->configService->get(self::JUNE_TRANSFER_FULFILLMENT)) {
                try {
                    $event_data = $event->getPayloads();
                    if (isset($event_data[0]["orderId"])) {
                        $orderId = $event_data[0]["orderId"];
                        $order = $this->getOrder($orderId, Context::createDefaultContext());
                        if ($order) {
                            $customer = $order->getOrderCustomer()?->getCustomer() ?? $order->getOrderCustomer();
                            $products = [];

                            $itemCriteria = new Criteria();
                            $itemCriteria->addFilter(new EqualsFilter("orderId", $order->getId()));
                            $lineItems = $this->orderLineItemsRepository->search($itemCriteria, Context::createDefaultContext())->getElements();

                            if ($order->getStateMachineState()->getTechnicalName() === "completed") {
                                $voucher = 0;
                                $shopUrl = $this->getShopUrl($order->getSalesChannelId(), $order->getLanguageId());
                                $orderDate = $order->getOrderDateTime()->format('c');

                                $city = "";

                                foreach ($lineItems as $item) {
                                    if ($product = $this->transformLineItem($item, $shopUrl, $order->getSalesChannelId(), $voucher)) {
                                        $products[] = $product;
                                    }
                                }

                                $data = [
                                    'action'    => 'order_fulfilled',
                                    'id'        => ($customer ? ($customer->getCustomerNumber() ?? $customer->getId()) : session_id()),
                                    'email'     => $customer?->getEmail(),
                                    'customer'  => [
                                        'firstname'     => $customer?->getFirstname(),
                                        'lastname'      => $customer?->getLastname(),
                                        'store_id'      => $order->getSalesChannelId(),
                                        'locale'        => $this->getLanguageCode($order->getLanguageId()),
                                        'geo_location'  => [
                                            'city'          => $city,
                                            'country'       => $this->getCountry($order->getSalesChannelId()),
                                        ],
                                        'guest'         => null
                                    ],
                                    'fulfillment' => [
                                        'id'            => $order->getOrderNumber(),
                                        'created_at'    => $orderDate,
                                        'fulfilled_at'  => null,
                                        'total'         => $order->getPrice()?->getTotalPrice(),
                                        'shipping'      => $order->getShippingCosts()?->getTotalPrice(),
                                        'voucher'       => $voucher,
                                        'products'      => $products,
                                    ],
                                ];

                                if ($order->getDeliveries()) {
                                    foreach ($order->getDeliveries() AS $delivery) {
                                        if ($delivery->getStateMachineState()->getTechnicalName() === 'shipped') {
                                            $data['fulfillment']['fulfilled_at'] = $delivery->getUpdatedAt()->format('c');
                                        }
                                    }
                                }

                                // set new updatedAt date for customer so all properties are exported on the next schedule run
                                if ($this->configService->get(self::JUNE_TRANSFER_CUSTOMERS)) {
                                    $this->createJuneCustomer($customer, $order->getSalesChannelId(), $this->getLanguageCode($order->getLanguageId()));

                                    // trigger update to make sure customer is exported in schedule
                                    $customerId = ($customer instanceof CustomerEntity ? $customer->getId() : $customer->getCustomerId());
                                    if ($customerId) {
                                        $this->customerEntity->update([
                                            [
                                                'id' => $customerId
                                            ]
                                        ], $event->getContext());
                                    }
                                }

                                $this->juneApiClient->sendData($data, "https://engagement.juneapp.com/commerce/webhook/$juneCommerceToken", null);
                            }
                        }
                    }
                } catch (\Exception $e) {
                    $this->loggerHelper->logError($e->getMessage(), "commerceOrderFulfilled");
                }
            }
        } else {
            $this->loggerHelper->logError("The Commerce Token in the JUNE Plugin configuration is missing", "pluginConfig");
        }
    }

    /**
     * @param OrderLineItemEntity|LineItem $item
     * @param $shopUrl
     * @param $salesChannelId
     * @param $voucher
     * @return array|void
     */
    private function transformLineItem(OrderLineItemEntity|LineItem $item, $shopUrl, $salesChannelId, &$voucher) {
        try {
            if ($item->getType() === "promotion") {
                $voucher = $voucher + round($item instanceof LineItem ? $item?->getPrice()?->getTotalPrice() : $item->getTotalPrice(), 2);
            }

            $productData = $item->getPayload();

            $categories = [];
            if (!empty($productData['categoryIds'])) {
                foreach ($productData['categoryIds'] as $categoryId) {
                    if ($categoryId) {
                        $categoryCriteria = new Criteria();
                        $categoryCriteria = $categoryCriteria->addFilter(new EqualsFilter('id', $categoryId));
                        $category = $this->categoryRepository->search($categoryCriteria, $this->createSalesChannelContext($salesChannelId))->first();
                        if ($category?->getName()) {
                            $categories[] = $category->getName();
                        }
                    }
                }
            }

            $productId = ($item instanceof LineItem ? $item->getId() : $item->getProductId());
            $productExternalIdentifier = !empty($productData['productNumber']) ? $productData['productNumber'] : $item->getId();
            $productPrice = ($item instanceof LineItem ? $item?->getPrice()?->getTotalPrice() : $item->getTotalPrice());

            // load variant definitions for cart items
            $productName = $item->getLabel();
            if (!empty($productData['options'])) {
                foreach ($productData['options'] as $optionItem) {
                    if (!empty($optionItem['option'])) {
                        $productName .= " | " . $optionItem['option'];
                    }
                }
            }

            $product = [
                'id' => $productExternalIdentifier,
                'name' => $productName,
                'qty' => $item->getQuantity(),
                'price' => $productPrice,
                'category' => $categories
            ];

            if ($item->getCover() != NULL) {
                $product['image'] = $item?->getCover()?->getUrl();
            }

            if ($shopUrl && $productId) {
                // $item->getProduct()->getSeoUrls();
                $product['link'] = $shopUrl . "/detail/" . $productId;
            }

            return $product;
        } catch (\Exception $e) {
            $this->loggerHelper->logError($e->getMessage(), "commerceTransformLineItems");
        }
    }

    /**
     * @param GuestCustomerRegisterEvent $event
     * @return void
     */
    public function guestRegister(GuestCustomerRegisterEvent $event) {
        if($this->configService->get(self::JUNE_TRANSFER_GUESTDATA)) {
            $this->createJuneCustomer($event->getCustomer(), $event->getSalesChannelId(), $this->getLanguageCode($event->getSalesChannelContext()->getLanguageId()));
        }
    }

    /**
     * @param CustomerRegisterEvent $event
     * @return void
     */
    public function customerRegister(CustomerRegisterEvent $event) {
        $this->createJuneCustomer($event->getCustomer(), $event->getSalesChannelId(), $this->getLanguageCode($event->getSalesChannelContext()->getLanguageId()));
    }

    /**
     * @param CustomerEntity|OrderCustomerEntity $customer
     * @param $salesChannelId
     * @param $locale
     * @return void
     */
    public function createJuneCustomer(CustomerEntity|OrderCustomerEntity $customer, $salesChannelId, $locale):void {
        if ($this->configService->get(self::JUNE_TRANSFER_CUSTOMERS)) {
            try {
                $commerceCollectToken = $this->configService->get(JuneCommerceSubscriber::JUNE_COMMERCE_COLLECT_TOKEN);
                $commerceCollectAuth = $this->configService->get(JuneCommerceSubscriber::JUNE_COMMERCE_COLLECT_AUTH);

                if ($customer->getEmail() && $commerceCollectToken && $commerceCollectAuth) {
                    $customerUpdateFields = [
                        "external_id",
                        "firstname",
                        "lastname",
                        "salutation",
                        "salutation_readable",
                        "locale",
                        "updated_at",
                        "store_id",
                        "guest"
                    ];

                    $subscriber = $this->getNewsletterRecipientByEmail($customer->getEmail());

                    $salutation = $customer->getSalutation();
                    if (!$salutation && $customer->getSalutationId()) {
                        $salutation = $this->getSalutation($customer->getSalutationId());
                    }

                    $data = [
                        "customer"  => [
                            "external_id"           => $customer->getCustomerNumber() ?? $customer->getId(),
                            "email"                 => $customer->getEmail(),
                            "firstname"             => $customer->getFirstName(),
                            "lastname"              => $customer->getLastName(),
                            "salutation"            => $salutation?->getSalutationKey(),
                            "salutation_readable"   => $salutation?->getLetterName(),
                            "store_id"              => $salesChannelId,
                            "locale"                => $locale,
                            "created_at"            => $customer->getCreatedAt()?->format('c') ?? $subscriber?->getCreatedAt()?->format('c'),
                            "updated_at"            => $customer->getUpdatedAt()?->format('c') ?? $subscriber?->getUpdatedAt()?->format('c'),
                            "guest"                 => true
                        ]
                    ];

                    if ($customer instanceof CustomerEntity) {
                        $data["customer"]["geo_location"] = [
                            "city"                  => $customer->getDefaultBillingAddress()?->getCity(),
                            "continent"             => "",
                            "country"               => $customer->getDefaultBillingAddress()?->getCountry()?->getName()
                        ];

                        $data["customer"]["guest"] = $customer->getGuest();

                        if ($subscriber?->getConfirmedAt()) {
                            $data['customer']['subscription_confirmed_at'] = $subscriber->getConfirmedAt()->format('c');
                            $customerUpdateFields[] = 'subscription_confirmed_at';
                        }
                        if ($customer->getBirthday()) {
                            $data['customer']['birthday'] = $customer->getBirthday()->format('c');
                            if (!in_array('birthday', $customerUpdateFields)) {
                                $customerUpdateFields[] = 'birthday';
                            }
                        }
                        if ($customer->getLastLogin()) {
                            $data['customer']['last_login'] = $customer->getLastLogin()->format('c');
                            if (!in_array('last_login', $customerUpdateFields)) {
                                $customerUpdateFields[] = 'last_login';
                            }
                        }

                        if ($customer->getLastOrderDate()) {
                            $data['customer']['last_order_date'] = $customer->getLastOrderDate()->format('c');
                            if (!in_array('last_order_date', $customerUpdateFields)) {
                                $customerUpdateFields[] = 'last_order_date';
                            }
                        }
                        if ($customer->getOrderCount()) {
                            $data['customer']['order_count'] = $customer->getOrderCount();
                            if (!in_array('order_count', $customerUpdateFields)) {
                                $customerUpdateFields[] = 'order_count';
                            }
                        }
                        if ($customer->getOrderTotalAmount()) {
                            $data['customer']['order_total_amount'] = $customer->getOrderTotalAmount();
                            if (!in_array('order_total_amount', $customerUpdateFields)) {
                                $customerUpdateFields[] = 'order_total_amount';
                            }
                        }

                        if ($customer->getCampaignCode()) {
                            $data['customer']['campaign_code'] = $customer->getCampaignCode();
                            if (!in_array('campaign_code', $customerUpdateFields)) {
                                $customerUpdateFields[] = 'campaign_code';
                            }
                        }
                        if ($customer->getAffiliateCode()) {
                            $data['customer']['affiliate_code'] = $customer->getAffiliateCode();
                            if (!in_array('affiliate_code', $customerUpdateFields)) {
                                $customerUpdateFields[] = 'affiliate_code';
                            }
                        }

                        if ($customer->getTagIds()) {
                            $data['customer']['_tags'] = [];
                            foreach ($customer->getTagIds() as $tagId) {
                                if ($tag = $this->getTag($tagId)) {
                                    $data['customer']['_tags'][] = $tag;
                                }
                            }
                        }
                    }

                    // add more information to customer if available
                    if ($customer->getCustomFields()) {
                        foreach ($customer->getCustomFields() as $fieldName => $fieldValue) {
                            if (!isset($data['customer'][$fieldName])) {
                                $transformedFieldName = ($data['customer']['guest'] ? 'guest_' : '') . $fieldName;
                                $data['customer'][$transformedFieldName] = $fieldValue;

                                if (!in_array($transformedFieldName, $customerUpdateFields)) {
                                    $customerUpdateFields[] = $transformedFieldName;
                                }
                            }
                        }
                    }

                    $data['customer']['_unsubscribed'] = !($subscriber?->getStatus() === "direct" || $subscriber?->getStatus() === "optIn");
                    if (!in_array('_unsubscribed', $customerUpdateFields)) {
                        $customerUpdateFields[] = '_unsubscribed';
                    }

                    $data['customer']['intent'] = $subscriber?->getStatus();
                    if (!in_array('intent', $customerUpdateFields)) {
                        $customerUpdateFields[] = 'intent';
                    }

                    if ($subscriber?->getCity()) {
                        $data['customer']['city'] = $subscriber->getCity();
                        $data['customer']['geo_location']['city'] = $subscriber->getCity();
                        if (!in_array('city', $customerUpdateFields)) {
                            $customerUpdateFields[] = 'city';
                        }
                    }
                    if ($subscriber?->getZipCode()) {
                        $data['customer']['zip_code'] = $subscriber->getZipCode();
                        if (!in_array('zip_code', $customerUpdateFields)) {
                            $customerUpdateFields[] = 'zip_code';
                        }
                    }

                    if ($subscriber?->getTags()) {
                        if (!isset($data['customer']['_tags'])) {
                            $data['customer']['_tags'] = [];
                        }
                        foreach ($subscriber->getTags() as $tag) {
                            if ($tag?->getName() && !in_array($tag->getName(), $data['customer']['_tags'])) {
                                $data['customer']['_tags'][] = $tag->getName();
                            }
                        }
                    }

                    /* dont use default import because it relies on queue times
                    $requestData = [
                        "action" => "customer_create",
                        "json" => [
                            $data['customer']
                        ],
                        "activate_trigger" => true,
                        "default_subscribe" => false,
                        "force_subscribe" => false,
                        "update_if_exists" => true,
                        "overwrite_tags" => true,
                        "update_fields" => $customerUpdateFields
                    ];
                    $this->juneApiClient->sendData($requestData, "https://engagement.juneapp.com/collect/$commerceCollectToken/import/json", $commerceCollectAuth);
                    */

                    $this->upsertCustomer($customer, $data['customer'], $commerceCollectToken, $commerceCollectAuth);
                }
            } catch (\Exception $e) {
                $this->loggerHelper->logError($e->getMessage(), "commerceCustomerCreate");
            }
        }
    }

    /**
     * @param NewsletterRegisterEvent $event
     * @return void
     */
    public function newsletterRegister(NewsletterRegisterEvent $event) {
        if ($this->configService->get(self::JUNE_TRANSFER_SUBSCRIBERS)) {
            try {
                $commerceCollectToken = $this->configService->get(self::JUNE_COMMERCE_COLLECT_TOKEN);
                $commerceCollectAuth = $this->configService->get(self::JUNE_COMMERCE_COLLECT_AUTH);

                // $newsletterRecipient = $event->getNewsletterRecipient(); // The event newsletterRecipient does not contain any tags or tagIds...
                $subscriber = $this->getNewsletterRecipient($event->getNewsletterRecipientId());

                if ($subscriber && $subscriber->getEmail() && !empty($commerceCollectToken) && !empty($commerceCollectAuth)) {
                    $customerUpdateFields = [
                        "firstname",
                        "lastname",
                        "locale",
                        "_unsubscribed"
                    ];

                    $customer = $this->getCustomerByEmail($subscriber->getEmail());

                    $enrichedCustomer = [
                        'email'         => $subscriber->getEmail(),
                        'firstname'     => $subscriber->getFirstname(),
                        'lastname'      => $subscriber->getLastname(),
                        'store_id'      => $event->getSalesChannelId(),
                        'locale'        => $this->getLanguageCode($event->getContext()->getLanguageId()),
                        'guest'         => $customer?->getGuest() ?? false,
                        '_unsubscribed' => true,
                        'geo_location'  => []
                    ];

                    // add more information to customer if available
                    if ($subscriber?->getConfirmedAt()) {
                        $enrichedCustomer['subscription_confirmed_at'] = $subscriber->getConfirmedAt()->format('c');
                    }

                    if ($customer?->getCustomFields()) {
                        foreach ($customer->getCustomFields() as $fieldName => $fieldValue) {
                            if (!isset($enrichedCustomer[$fieldName])) {
                                $transformedFieldName = ($enrichedCustomer['guest'] ? 'guest_' : '') . $fieldName;
                                $enrichedCustomer[$transformedFieldName] = $fieldValue;

                                if (!in_array($transformedFieldName, $customerUpdateFields)) {
                                    $customerUpdateFields[] = $transformedFieldName;
                                }
                            }
                        }
                    }
                    if ($customer?->getBirthday()) {
                        $enrichedCustomer['birthday'] = $customer->getBirthday()->format('c');
                        if (!in_array('birthday', $customerUpdateFields)) {
                            $customerUpdateFields[] = 'birthday';
                        }
                    }
                    if ($customer?->getLastLogin()) {
                        $enrichedCustomer['last_login'] = $customer->getLastLogin()->format('c');
                        if (!in_array('last_login', $customerUpdateFields)) {
                            $customerUpdateFields[] = 'last_login';
                        }
                    }
                    if ($customer?->getLastOrderDate()) {
                        $enrichedCustomer['last_order_date'] = $customer->getLastOrderDate()->format('c');
                        if (!in_array('last_order_date', $customerUpdateFields)) {
                            $customerUpdateFields[] = 'last_order_date';
                        }
                    }
                    if ($customer?->getOrderCount()) {
                        $enrichedCustomer['order_count'] = $customer->getOrderCount();
                        if (!in_array('order_count', $customerUpdateFields)) {
                            $customerUpdateFields[] = 'order_count';
                        }
                    }
                    if ($customer?->getOrderTotalAmount()) {
                        $enrichedCustomer['order_total_amount'] = $customer->getOrderTotalAmount();
                        if (!in_array('order_total_amount', $customerUpdateFields)) {
                            $customerUpdateFields[] = 'order_total_amount';
                        }
                    }

                    if ($customer?->getCampaignCode()) {
                        $enrichedCustomer['campaign_code'] = $customer->getCampaignCode();
                        if (!in_array('campaign_code', $customerUpdateFields)) {
                            $customerUpdateFields[] = 'campaign_code';
                        }
                    }

                    if ($customer?->getAffiliateCode()) {
                        $enrichedCustomer['affiliate_code'] = $customer->getAffiliateCode();
                        if (!in_array('affiliate_code', $customerUpdateFields)) {
                            $customerUpdateFields[] = 'affiliate_code';
                        }
                    }

                    $enrichedCustomer['_unsubscribed'] = !($subscriber?->getStatus() === "direct" || $subscriber?->getStatus() === "optIn");
                    if (!in_array('_unsubscribed', $customerUpdateFields)) {
                        $customerUpdateFields[] = '_unsubscribed';
                    }

                    $enrichedCustomer['intent'] = $subscriber?->getStatus();
                    if (!in_array('intent', $customerUpdateFields)) {
                        $customerUpdateFields[] = 'intent';
                    }

                    if ($subscriber?->getCity()) {
                        $enrichedCustomer['city'] = $subscriber->getCity();
                        $enrichedCustomer['geo_location']['city'] = $subscriber->getCity();
                        if (!in_array('city', $customerUpdateFields)) {
                            $customerUpdateFields[] = 'city';
                        }
                    }
                    if ($subscriber?->getZipCode()) {
                        $enrichedCustomer['zip_code'] = $subscriber->getZipCode();
                        if (!in_array('zip_code', $customerUpdateFields)) {
                            $customerUpdateFields[] = 'zip_code';
                        }
                    }

                    if ($customer?->getTagIds()) {
                        $enrichedCustomer['_tags'] = [];
                        foreach ($customer->getTagIds() as $tagId) {
                            if ($tag = $this->getTag($tagId)) {
                                $enrichedCustomer['_tags'][] = $tag;
                            }
                        }
                    }
                    if ($subscriber?->getTags()) {
                        if (!isset($enrichedCustomer['_tags'])) {
                            $enrichedCustomer['_tags'] = [];
                        }
                        foreach ($subscriber->getTags() as $tag) {
                            if ($tag?->getName() && !in_array($tag->getName(), $enrichedCustomer['_tags'])) {
                                $enrichedCustomer['_tags'][] = $tag->getName();
                            }
                        }
                    }

                    /* dont use default import because it relies on queue times
                    $requestData = [
                        "action" => "subscription_create",
                        "json" => [$enrichedCustomer],
                        "activate_trigger" => true,
                        "default_subscribe" => false,
                        "force_subscribe" => false,
                        "update_if_exists" => true,
                        "update_fields" => $customerUpdateFields
                    ];
                    $this->juneApiClient->sendData($requestData, "https://engagement.juneapp.com/collect/$commerceCollectToken/import/json", $commerceCollectAuth);
                    */
                    $this->upsertCustomer($customer, $enrichedCustomer, $commerceCollectToken, $commerceCollectAuth);
                }
            } catch (\Exception $e) {
                $this->loggerHelper->logError($e->getMessage(), "commerceNewsletterRegister");
            }
        }
    }

    public function newsletterSubscribe(NewsletterConfirmEvent $event) {
        if ($this->configService->get(self::JUNE_TRANSFER_SUBSCRIBERS)) {
            try {
                $commerceCollectToken = $this->configService->get(self::JUNE_COMMERCE_COLLECT_TOKEN);
                $commerceCollectAuth = $this->configService->get(self::JUNE_COMMERCE_COLLECT_AUTH);

                // $subscriber = $event->getNewsletterRecipient(); // The event newsletterRecipient does not contain any tags or tagIds...
                $subscriber = $this->getNewsletterRecipient($event->getNewsletterRecipientId());

                if ($subscriber && $subscriber->getEmail() && !empty($commerceCollectToken) && !empty($commerceCollectAuth)) {
                    $customerUpdateFields = [
                        "firstname",
                        "lastname",
                        "locale",
                        "_unsubscribed"
                    ];

                    $customer = $this->getCustomerByEmail($subscriber->getEmail());

                    $enrichedCustomer = [
                        'email'         => $subscriber->getEmail(),
                        'firstname'     => $subscriber->getFirstname(),
                        'lastname'      => $subscriber->getLastname(),
                        'store_id'      => $event->getSalesChannelId(),
                        'locale'        => $this->getLanguageCode($event->getContext()->getLanguageId()),
                        'guest'         => $customer?->getGuest() ?? false,
                        '_unsubscribed' => false,
                        'geo_location'  => []
                    ];

                    // add more information to customer if available
                    if ($subscriber?->getConfirmedAt()) {
                        $enrichedCustomer['subscription_confirmed_at'] = $subscriber->getConfirmedAt()->format('c');
                    }

                    if ($customer?->getCustomFields()) {
                        foreach ($customer->getCustomFields() as $fieldName => $fieldValue) {
                            if (!isset($enrichedCustomer[$fieldName])) {
                                $transformedFieldName = ($enrichedCustomer['guest'] ? 'guest_' : '') . $fieldName;
                                $enrichedCustomer[$transformedFieldName] = $fieldValue;

                                if (!in_array($transformedFieldName, $customerUpdateFields)) {
                                    $customerUpdateFields[] = $transformedFieldName;
                                }
                            }
                        }
                    }
                    if ($customer?->getBirthday()) {
                        $enrichedCustomer['birthday'] = $customer->getBirthday()->format('c');
                        if (!in_array('birthday', $customerUpdateFields)) {
                            $customerUpdateFields[] = 'birthday';
                        }
                    }
                    if ($customer?->getLastLogin()) {
                        $enrichedCustomer['last_login'] = $customer->getLastLogin()->format('c');
                        if (!in_array('last_login', $customerUpdateFields)) {
                            $customerUpdateFields[] = 'last_login';
                        }
                    }
                    if ($customer?->getLastOrderDate()) {
                        $enrichedCustomer['last_order_date'] = $customer->getLastOrderDate()->format('c');
                        if (!in_array('last_order_date', $customerUpdateFields)) {
                            $customerUpdateFields[] = 'last_order_date';
                        }
                    }
                    if ($customer?->getOrderCount()) {
                        $enrichedCustomer['order_count'] = $customer->getOrderCount();
                        if (!in_array('order_count', $customerUpdateFields)) {
                            $customerUpdateFields[] = 'order_count';
                        }
                    }
                    if ($customer?->getOrderTotalAmount()) {
                        $enrichedCustomer['order_total_amount'] = $customer->getOrderTotalAmount();
                        if (!in_array('order_total_amount', $customerUpdateFields)) {
                            $customerUpdateFields[] = 'order_total_amount';
                        }
                    }

                    if ($customer?->getCampaignCode()) {
                        $enrichedCustomer['campaign_code'] = $customer->getCampaignCode();
                        if (!in_array('campaign_code', $customerUpdateFields)) {
                            $customerUpdateFields[] = 'campaign_code';
                        }
                    }

                    if ($customer?->getAffiliateCode()) {
                        $enrichedCustomer['affiliate_code'] = $customer->getAffiliateCode();
                        if (!in_array('affiliate_code', $customerUpdateFields)) {
                            $customerUpdateFields[] = 'affiliate_code';
                        }
                    }

                    $enrichedCustomer['_unsubscribed'] = !($subscriber?->getStatus() === "direct" || $subscriber?->getStatus() === "optIn");
                    if (!in_array('_unsubscribed', $customerUpdateFields)) {
                        $customerUpdateFields[] = '_unsubscribed';
                    }

                    $enrichedCustomer['intent'] = $subscriber?->getStatus();
                    if (!in_array('intent', $customerUpdateFields)) {
                        $customerUpdateFields[] = 'intent';
                    }

                    if ($subscriber?->getCity()) {
                        $enrichedCustomer['city'] = $subscriber->getCity();
                        $enrichedCustomer['geo_location']['city'] = $subscriber->getCity();
                        if (!in_array('city', $customerUpdateFields)) {
                            $customerUpdateFields[] = 'city';
                        }
                    }
                    if ($subscriber?->getZipCode()) {
                        $enrichedCustomer['zip_code'] = $subscriber->getZipCode();
                        if (!in_array('zip_code', $customerUpdateFields)) {
                            $customerUpdateFields[] = 'zip_code';
                        }
                    }

                    if ($customer?->getTagIds()) {
                        $enrichedCustomer['_tags'] = [];
                        foreach ($customer->getTagIds() as $tagId) {
                            if ($tag = $this->getTag($tagId)) {
                                $enrichedCustomer['_tags'][] = $tag;
                            }
                        }
                    }
                    if ($subscriber?->getTags()) {
                        if (!isset($enrichedCustomer['_tags'])) {
                            $enrichedCustomer['_tags'] = [];
                        }
                        foreach ($subscriber->getTags() as $tag) {
                            if ($tag?->getName() && !in_array($tag->getName(), $enrichedCustomer['_tags'])) {
                                $enrichedCustomer['_tags'][] = $tag->getName();
                            }
                        }
                    }

                    /* dont use default import because it relies on queue times
                    $requestData = [
                        "action" => "subscription_subscribe",
                        "json" => [$enrichedCustomer],
                        "activate_trigger" => true,
                        "default_subscribe" => true,
                        "force_subscribe" => true,
                        "update_if_exists" => true,
                        "overwrite_tags" => true,
                        "update_fields" => $customerUpdateFields
                    ];
                    $this->juneApiClient->sendData($requestData, "https://engagement.juneapp.com/collect/$commerceCollectToken/import/json", $commerceCollectAuth);
                    */
                    $this->upsertCustomer($customer, $enrichedCustomer, $commerceCollectToken, $commerceCollectAuth);
                }
            } catch (\Exception $e) {
                $this->loggerHelper->logError($e->getMessage(), "commerceNewsletterSubscribe");
            }
        }
    }

    public function newsletterUnsubscribe(NewsletterUnsubscribeEvent|EntityDeletedEvent $event) {
        if ($this->configService->get(self::JUNE_TRANSFER_SUBSCRIBERS)) {
            try {
                $commerceCollectToken = $this->configService->get(self::JUNE_COMMERCE_COLLECT_TOKEN);
                $commerceCollectAuth = $this->configService->get(self::JUNE_COMMERCE_COLLECT_AUTH);

                if (!empty($commerceCollectToken) && !empty($commerceCollectAuth)) {
                    $updateFields = [
                        "firstname",
                        "lastname",
                        "locale",
                        "_unsubscribed"
                    ];

                    // $newsletterRecipient = $event->getNewsletterRecipient(); // The event newsletterRecipient does not contain any tags or tagIds...
                    $newsletterRecipient = $this->getNewsletterRecipient($event->getNewsletterRecipientId());
                    $customer = $this->getCustomerByEmail($newsletterRecipient->getEmail());
                    $enrichedCustomer = [
                        'email'         => $newsletterRecipient->getEmail(),
                        'firstname'     => $newsletterRecipient->getFirstname(),
                        'lastname'      => $newsletterRecipient->getLastname(),
                        'store_id'      => $event->getSalesChannelId(),
                        'locale'        => $this->getLanguageCode($event->getContext()->getLanguageId()),
                        '_unsubscribed' => true,
                        'geo_location'  => []
                    ];

                    if ($newsletterRecipient->getCity()) {
                        $updateFields[] = 'city';
                        $enrichedCustomer['city'] = $newsletterRecipient->getCity();
                        $enrichedCustomer['geo_location']['city'] = $enrichedCustomer['city'];
                    }
                    if ($newsletterRecipient->getZipCode()) {
                        $updateFields[] = 'zip_code';
                        $enrichedCustomer['zip_code'] = $newsletterRecipient->getZipCode();
                    }

                    if ($newsletterRecipient?->getTags()) {
                        $enrichedCustomer['_tags'] = [];
                        foreach ($newsletterRecipient->getTags() as $tag) {
                            if ($tag?->getName()) {
                                $enrichedCustomer['_tags'][] = $tag->getName();
                            }
                        }
                    }

                    /* dont use default import because it relies on queue times
                    $requestData = [
                        "action" => "subscription_unsubscribe",
                        "json" => [$enrichedCustomer],
                        "activate_trigger" => true,
                        "default_subscribe" => false,
                        "force_subscribe" => false,
                        "update_if_exists" => true,
                        "update_fields" => $updateFields
                    ];
                    $this->juneApiClient->sendData($requestData, "https://engagement.juneapp.com/collect/$commerceCollectToken/import/json", $commerceCollectAuth);
                    */
                    $this->upsertCustomer($customer, $enrichedCustomer, $commerceCollectToken, $commerceCollectAuth);
                }
            } catch (\Exception $e) {
                $this->loggerHelper->logError($e->getMessage(), "commerceNewsletterUnsubscribe");
            }
        }
    }

    private function upsertCustomer($customer, $collectData, $collectToken, $collectAuth) {
        // load JuneId if available
        $collectId = $customer?->getExtension('juneCustomer')?->getVars()['juneId'];

        $customerResponse = $this->juneApiClient->upsertCustomer($collectData, $collectId, $collectToken, $collectAuth);

        // update customer so it is checked in the scheduled export
        if ($customer) {
            $customerUpdateData = [
                'id' => $customer->getId()
            ];
            if (!$customer?->getExtension('juneCustomer')?->getVars()['juneId'] && $customerResponse && $customerResponse['id']) {
                // add juneId if necessary
                $customerUpdateData['juneCustomer'] = [
                    'juneId' => $customerResponse['id']
                ];
            }
            $this->customerEntity->update([
                $customerUpdateData
            ], Context::createDefaultContext());
        }
    }

    public function createSalesChannelContext(string $salesChannelId): SalesChannelContext
    {
        return $this->salesChannelContextFactory->create('', $salesChannelId, []);
    }

    public function getShopUrl($salesChannelId, $languageId) {
        $salesChannelUrls = [];
        $salesCriteria = new Criteria();
        $salesCriteria = $salesCriteria->addAssociation('domains');
        $salesChannelIds = $this->salesRepository->search($salesCriteria, Context::createDefaultContext());
        foreach($salesChannelIds->getEntities()->getElements() as $key => $salesChannel){
            foreach($salesChannel->getDomains() as $element) {
                array_push($salesChannelUrls, $element);
            }
        }
        foreach ($salesChannelUrls as $url) {
            if($url->getSalesChannelId() === $salesChannelId && $url->getLanguageId() === $languageId) {
                return $url->getUrl();
            }
        }
        return null;
    }

    public function getLanguageCode($languageId) {
        $languageCriteria = new Criteria();
        $languageCriteria = $languageCriteria->addFilter(new EqualsFilter('id', $languageId));
        $language = $this->languageRepository->search($languageCriteria, Context::createDefaultContext())->first();

        $localeCriteria = new Criteria();
        $localeCriteria = $localeCriteria->addFilter(new EqualsFilter('id', $language->getLocaleId()));
        $locale = $this->localeRepository->search($localeCriteria, Context::createDefaultContext())->first();

        return $locale->getCode();
    }

    public function getCountry($salesChannelId) {
        $salesChannelCriteria = new Criteria();
        $salesChannelCriteria = $salesChannelCriteria->addFilter(new EqualsFilter('id', $salesChannelId));
        $salesChannel = $this->salesRepository->search($salesChannelCriteria, Context::createDefaultContext())->first();

        $countryCriteria = new Criteria();
        $countryCriteria = $countryCriteria->addFilter(new EqualsFilter('id', $salesChannel->getCountryId()));
        $country = $this->countryRepository->search($countryCriteria, Context::createDefaultContext())->first();

        return $country->getName();
    }

    private function getOrder(string $orderId, Context $context): ?OrderEntity
    {
        $orderCriteria = new Criteria();
        $orderCriteria = $orderCriteria->addFilter(new EqualsFilter('id', $orderId));
        $orderCriteria->addAssociation('deliveries');

        return $this->orderRepository->search($orderCriteria, $context)->first();
    }

    private function getCustomer($customerId): ?CustomerEntity {
        if ($customerId) {
            $customerSearch = new Criteria([$customerId]);
            $customerSearch->addAssociation('tags');
            return $this->customerEntity->search($customerSearch, Context::createDefaultContext())->first();
        }
        return null;
    }

    private function getCustomerByEmail(string $email): ?CustomerEntity {
        if ($email) {
            $customerSearch = new Criteria();
            $customerSearch->addFilter(new EqualsFilter('email', $email));
            $customerSearch->addAssociation('tags');
            return $this->customerEntity->search($customerSearch, Context::createDefaultContext())->first();
        }
        return null;
    }

    private function getNewsletterRecipient($recipientId): ?NewsletterRecipientEntity {
        if ($recipientId) {
            $recipientSearch = new Criteria([$recipientId]);
            $recipientSearch->addAssociation('tags');
            return $this->newsletterRecipientRepository->search($recipientSearch, Context::createDefaultContext())->first();
        }
        return null;
    }

    private function getNewsletterRecipientByEmail(string $email): ?NewsletterRecipientEntity {
        if ($email) {
            $recipientSearch = new Criteria();
            $recipientSearch->addFilter(new EqualsFilter('email', $email));
            $recipientSearch->addAssociation('tags');
            return $this->newsletterRecipientRepository->search($recipientSearch, Context::createDefaultContext())->first();
        }
        return null;
    }

    private function getTag(string $tagId) {
        if (!in_array($tagId, $this->loadedCustomerTags)) {
            $tag = $this->customerTagsRepository->search(new Criteria([$tagId]), Context::createDefaultContext())->first();
            $this->loadedCustomerTags[$tagId] = $tag->getName();
        }

        return $this->loadedCustomerTags[$tagId];
    }

    private function getSalutation($salutationId) {
        if ($salutationId) {
            $salutationSearch = new Criteria([$salutationId]);
            return $this->salutationRepository->search($salutationSearch, Context::createDefaultContext())->first();
        }
        return null;
    }
}
